import os
from dataclasses import dataclass, field
import yaml
from typing import Dict, Any, Optional
import string
import numpy as np
import healpy as hp
import sys
import argparse
import yaml


def join(loader: yaml.Loader, node: yaml.Node) -> str:
    """
    Custom YAML tag to join sequences into a single string.

    Parameters
    ----------
        loader : yaml.Loader
            The YAML loader instance.
        node : yaml.Node
            The YAML node to process.

    Returns
    -------
        str
            A string formed by joining the sequence elements.
    """
    seq = loader.construct_sequence(node)
    return "".join([str(i) for i in seq])


def parse_args() -> argparse.Namespace:
    """
    Parse command-line arguments, especially the configuration file path.

    Parameters
    ----------
        None

    Returns
    -------
        argparse.Namespace
            The parsed arguments.
    """
    parser = argparse.ArgumentParser(
        description="Parse arguments and return parser object"
    )

    parser.add_argument(
        "-C",
        "--config_path",
        type=str,
        default="configs/test_config.yaml",
        help="Path to the configuration file. If not provided, takes the default path.",
    )

    return parser.parse_args()

def get_params(config_path: Optional[str] = None) -> "Configs":
    """
    Get parameters from a YAML configuration file.

    Parameters
    ----------
        config_path : str, optional
            Path to the configuration file. If not provided, uses the default path.

    Returns
    -------
        Configs
            An instance of the Configs class containing the configuration parameters.
    """
    
    yaml.add_constructor("!join", join)
    if config_path is None:
        parsed_args = parse_args()
        config_path = os.path.abspath(parsed_args.config_path)
    return Configs(config_path=config_path)

@dataclass
class InstrumentConfig:
    #frequency: list = field(default_factory=list)
    #depth_I: list = field(default_factory=list)
    #depth_P: list = field(default_factory=list)
    #fwhm: list = field(default_factory=list)
    #bandwidth: list = field(default_factory=list)
    #ell_knee: list = field(default_factory=list)
    #alpha_knee: list = field(default_factory=list)
    #channels_tags: list = field(default_factory=list)

    def load_from_yaml(self, yaml_data: Dict[str, Any], experiment: str):
        experiment_data = yaml_data.get(experiment, {})

        attrs = [
            'frequency', 'depth_I', 'depth_P', 'fwhm', 'bandwidth',
            'path_bandpasses', 'ell_knee', 'alpha_knee',
            'path_hits_maps', 'path_depth_maps'
        ]

        for attr in attrs:
            if attr in experiment_data:
                setattr(self, attr, experiment_data[attr])

        self.beams = experiment_data.get("beams", "gaussian")
        if self.beams != "gaussian":
            if 'path_beams' not in experiment_data:
                raise ValueError("Missing 'path_beams' for non-gaussian beams in the experiment yaml file.")
            self.path_beams = experiment_data['path_beams']
        else:
            if 'fwhm' not in experiment_data:
                raise ValueError(f"FWHM must be provided in the yaml file for gaussian beams.")

#        for attr in ['path_bandpasses', 'path_hits_maps', 'path_depth_maps', 'path_beams']:
#            if hasattr(self, attr):
#                setattr(self, attr, os.path.join(path_utils, getattr(self, attr)))

        self.channels_tags = experiment_data.get("channels_tags", [])
        if not self.channels_tags:
            self._generate_channel_tags()

    def _generate_channel_tags(self):
        self.channels_tags = []
        unique_freqs, counts = np.unique(self.frequency, return_counts=True)
        labels = [list(string.ascii_lowercase[:c]) for c in counts]

        for freq in self.frequency:
            idx = np.where(unique_freqs == freq)[0][0]
            label = labels[idx].pop(0)
            tag = f"{freq}GHz" if counts[idx] == 1 else f"{freq}{label}GHz"
            self.channels_tags.append(tag)
        
@dataclass
class Configs:
    """
    Class to store settings and relevant quantities for the main script.
    """

    config_path: Optional[str] = None
    config: Optional[Dict[str, Any]] = field(default_factory=dict)

    def __post_init__(self):
        if self.config_path:
            self.config = yaml.load(open(self.config_path), Loader=yaml.FullLoader)
        self._store_passed_settings()
        if self.experiment:
            self.instrument = InstrumentConfig()
            self._load_experiment_parameters()
            if self.generate_input_foregrounds and self.bandpass_integrate:
                    if not hasattr(self.instrument, 'bandwidth') and not hasattr(self.instrument, 'path_bandpasses'):
                        raise ValueError(f"If bandpass_integrate is True, 'bandwidth' or 'path_bandpasses' must be provided in the experiment yaml file.")
            self.bring_to_common_resolution = self.config.get("bring_to_common_resolution", True)
        

    def _store_passed_settings(self):
        self.nside = self.config["nside"]
        self.data_type = self.config["data_type"]

#        base_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), '..'))
        base_dir = os.path.abspath(os.path.dirname(__file__))

        self.lmin = self.config.get("lmin") or 2    
        self.lmax = self.config.get("lmax") or 2 * self.nside
        self.nside_in = self.config.get("nside_in") or self.nside
        self.fwhm_out = self.config.get("fwhm_out") or 2.5 * hp.nside2resol(self.nside, arcmin=True)
        self.verbose = self.config.get("verbose", False)
        self.nsim_start = self.config.get("nsim_start") or 0
        self.nsims = self.config.get("nsims") or 1
        self.parallelize = self.config.get("parallelize", False)
#        self.path_utils = self.config.get("path_utils") or os.path.join(base_dir, "utils")
        self.compsep = self.config.get("compsep") or ""
        self.compsep_residuals = self.config.get("compsep_residuals") or ""
        self.real_mc_tracers = self.config.get("real_mc_tracers") or ""
        self.combine_outputs = self.config.get("combine_outputs") or ""
        self.foreground_models = self.config.get("foreground_models") or ["d0","s0"]
        self.field_in = self.config.get("field_in") or "TQU"
        self.field_out = self.config.get("field_out") or "TQU"
        self.experiments_file = self.config.get("experiments_file") or os.path.join(base_dir, "utils", "experiments.yaml") 
        self.experiment = self.config.get("experiment") or ""
        self.pixel_window_in = self.config.get("pixel_window_in", False)
        self.pixel_window_out = self.config.get("pixel_window_out", False)
        self.units = self.config.get("units") or "uK_CMB"
        self.coordinates = self.config.get("coordinates") or "G"
#        self.mask_path = self.config.get("mask_path")
#        self.mask_type = self.config.get("mask_type") or "mask_for_compsep"
        self.mask_observations = self.config.get("mask_observations", None)
        self.mask_covariance = self.config.get("mask_covariance", None)
        self.leakage_correction = self.config.get("leakage_correction", None)
        if self.compsep or self.compsep_residuals or self.combine_outputs:
            self.save_compsep_products = self.config.get("save_compsep_products", True)
            self.return_compsep_products = self.config.get("return_compsep_products", False)
            if not self.save_compsep_products and not self.return_compsep_products:
                raise ValueError("At least one of save_compsep_products and return_compsep_products must be True.")
        self.path_outputs = self.config.get("path_outputs") or os.path.join(os.getcwd(), "outputs", self.experiment, ''.join(self.foreground_models))
#        self.path_outputs = os.path.normpath(os.path.join(os.getcwd(), self.path_outputs))

        self.generate_input_foregrounds = self.config.get("generate_input_foregrounds", True)
        self.generate_input_noise = self.config.get("generate_input_noise", True)
        self.generate_input_cmb = self.config.get("generate_input_cmb", True)
        self.generate_input_data = self.config.get("generate_input_data", True)
        self.bandpass_integrate = self.config.get("bandpass_integrate", False)
        if self.generate_input_foregrounds or self.generate_input_noise or self.generate_input_cmb or self.generate_input_data:
            self.save_inputs = self.config.get("save_inputs", False)
        if self.generate_input_noise:
            self.seed_noise = self.config.get("seed_noise", None)
        if self.generate_input_cmb:
            self.seed_cmb = self.config.get("seed_cmb", None)
        self.cls_cmb_path = self.config.get("cls_cmb_path") or os.path.join(base_dir, "utils", "Cls_Planck2018_lensed_r0.fits")
        self.cls_cmb_new_ordered = self.config.get("cls_cmb_new_ordered", True)

        # Input/output paths   
        dataname = f"total_{self.data_type}_ns{self.nside}_lmax{self.lmax}"
        def_data_path = os.path.join(os.getcwd(), "inputs", self.experiment, "total", ''.join(self.foreground_models), dataname)
        self.data_path = self.config.get("data_path") or def_data_path
#        self.data_path = os.path.normpath(os.path.join(os.getcwd(), self.data_path))

        noisename = f"noise_{self.data_type}_ns{self.nside}_lmax{self.lmax}"
        def_noise_path = os.path.join(os.getcwd(), "inputs", self.experiment, "noise", noisename)
        self.noise_path = self.config.get("noise_path") or def_noise_path
#        self.noise_path = os.path.normpath(os.path.join(os.getcwd(), self.noise_path))

        cmbname = f"cmb_{self.data_type}_ns{self.nside}_lmax{self.lmax}"
        def_cmb_path = os.path.join(os.getcwd(), "inputs", self.experiment, "cmb", cmbname)
        self.cmb_path = self.config.get("cmb_path") or def_cmb_path
#        self.cmb_path = os.path.normpath(os.path.join(os.getcwd(), self.cmb_path))

        fgdsname = f"foregrounds_{self.data_type}_ns{self.nside}_lmax{self.lmax}"
        def_fgds_path = os.path.join(os.getcwd(), "inputs", self.experiment, "foregrounds", ''.join(self.foreground_models), fgdsname)
        self.fgds_path = self.config.get("fgds_path") or def_fgds_path
#        self.fgds_path = os.path.normpath(os.path.join(os.getcwd(), self.fgds_path))

        self.return_fgd_components = self.config.get("return_fgd_components", False)

        self._validate_paths()

        self.compute_spectra = self.config.get("compute_spectra") or ""
        if self.compute_spectra:
            self.delta_ell = self.config.get("delta_ell") or 1
            self.spectra_comp = self.config.get("spectra_comp") or "anafast"
            self.return_Dell = self.config.get("return_Dell", False)
            self.field_cls_out = self.config.get("field_cls_out") or self.field_out
            self.save_spectra = self.config.get("save_spectra", True)
            self.return_spectra = self.config.get("return_spectra", True)
            self.save_mask = self.config.get("save_mask", True)
            
    def _validate_paths(self):
        for name, flag in zip(
            ['fgds_path', 'noise_path', 'cmb_path', 'data_path'],
            [self.generate_input_foregrounds, self.generate_input_noise, self.generate_input_cmb, self.generate_input_data]
        ):
            if ((not flag) or self.save_inputs) and not getattr(self, name):
                raise ValueError(f"Path '{name}' must be specified.")

    def _load_experiment_parameters(self):
        experiments_yaml_path = self.experiments_file
        if os.path.exists(experiments_yaml_path):
            with open(experiments_yaml_path, 'r') as file:
                experiments_data = yaml.safe_load(file)
                self.instrument.load_from_yaml(experiments_data, self.experiment)

    def to_dict_for_mc(self) -> Dict[str, Any]:
        """
        Extract relevant attributes as a dictionary.
        """
        return {
            'lmin': self.lmin,
            'lmax': self.lmax,
            'nside': self.nside,
            'data_type': self.data_type,
            'fwhm_out': self.fwhm_out,
            'foreground_models': self.foreground_models,
            'experiment': self.experiment,
            'pixel_window_in': self.pixel_window_in,
            'pixel_window_out': self.pixel_window_out,
            'units': self.units,
            'coordinates': self.coordinates,
            'bandpass_integrate': self.bandpass_integrate,
#            'mask_path': self.mask_path,
            'mask_observations': self.mask_observations,
            'mask_covariance': self.mask_covariance,
            'verbose': self.verbose,
            'cls_cmb_path': self.cls_cmb_path,
            'cls_cmb_new_ordered': self.cls_cmb_new_ordered,
        }


__all__ = [
    name
    for name, obj in globals().items()
    if callable(obj) and getattr(obj, "__module__", None) == __name__
] 


import numpy as np 
import healpy as hp
import os
from .routines import _slice_data, _map2alm_kwargs, _log
from .needlets import _get_needlet_windows_,  _needlet_filtering
from .configurations import Configs
from types import SimpleNamespace
from .simulations import _get_data_foregrounds_, _get_data_simulations_
from typing import Optional, Union, List, Dict, Any, Tuple
import sys


def get_and_save_real_tracers_B(
    config: Configs,
    foregrounds: Optional[np.ndarray] = None,
    systematics: Optional[np.ndarray] = None,
    **kwargs: Any
) -> None:
    """
    Generate and save realistic tracer for MC-ILC given the instrumental and parameters configuration.

    Parameters
    ----------
        config : Configs
            Configuration object containing the instrumental and parameters configuration. 
            It should contain the following attributes:
            - experiment : str
                Name of the experiment
            - lmax : int 
                Maximum multipole for the simulations and analysis.      
            - lmin : int
                Minimum multipole for the simulations and analysis.
            - nside : int
                HEALPix resolution parameter.   
            - data_type : str
                Type of data to be used in the simulations, either "maps" or "alms". 
                It also identifies the type of data associated eventually to provided foregrounds and systematics.
            - fwhm_out : float
                Full width at half maximum of the output beam in arcminutes used to generate the tracer.
            - foreground_models : list
                List of foreground models to be used in the simulations. Used if foregrounds is None.
            - real_mc_tracers : dict
                List of dictionaries containing the information about the MC-ILC tracers to be generated.
                It should contain the following keys:
                   - channels_tracers: list
                        List of channels indexes to be used for the tracers generation. 
                   - path_tracers : str
                        Path where the tracers will be saved.
            - bandpass_integrate : bool
                Boolean indicating whether to integrate over the bandpass of the instrument in the simulations.
            - instrument : dict
                Instrument configuration object containing the following attributes:
                   - channels_tags: list
                        List of tags for the instrument channels.
            - mask_observations : str
                Path to the mask which sets to zero unobserved pixels. If None, no mask is applied.
            - coordinates : str
                Coordinate system used for the simulations, either "G" for Galactic, "E" for Ecliptic, or "C" for Celestial.
            - pixel_window_in : bool
                Boolean indicating whether to apply the pixel window function to the input data.
            - pixel_window_out : str 
                Boolean indicating whether to apply the pixel window function to the output tracer.
            - verbose : bool
                Boolean indicating whether to print information about the process.
            
        foregrounds : np.ndarray, optional
            Foregrounds data to be used in the simulations for tracer generation. If None, the foregrounds will be generated using the configuration parameters.
            Shape depends on `config.data_type`:
            - If "maps": (n_channels, 3, n_pixels) for (T, Q, U)
            - If "alms": (n_channels, 3, n_alms) for (T, E, B) 
        
            
        systematics : np.ndarray, optional
            Systematics data to be added to the simulations for tracer generation. If None, no systematics are added.
            If provided, it should be a 3D array with shape: 
        (n_Shape depends on `config.data_type`:
            - If "maps": (n_channels, 3, n_pixels) for (T, Q, U)
            - If "alms": (n_channels, 3, n_alms) for (T, E, B) 

        kwargs : dict, optional
            Additional keyword arguments to be passed to healpy 'map2alm' function.

    Returns
    -------
        None
            Saves the generated tracers to the configured output path.
    """
    from broom import component_separation
    if "tracers_inputs_path" not in config.real_mc_tracers[0]:
        config.real_mc_tracers[0]["tracers_inputs_path"] = f"inputs_mc_tracers/{config.experiment}"

    config_mc = get_mc_config(config, config.real_mc_tracers[0]["tracers_inputs_path"])
    kwargs = _map2alm_kwargs(**kwargs)

    _log("Generating input simulations for MC-ILC tracers", verbose=config_mc.verbose)

    mc_data = get_mc_data(config_mc, foregrounds=foregrounds, systematics=systematics, **kwargs)

#    if systematics is not None:
#        print("Adding systematic effect to the data")
#        systematics = _adapt_systamatics(
#            systematics, mc_data.total.shape, config.lmax,
#            data_type=config_mc.data_type, **kwargs
#        )
#        mc_data.total = mc_data.total + systematics

    config_mc.compsep = get_tracers_compsep(config.real_mc_tracers[0]["channels_tracers"], config.lmax)

    _log(f"Generating the MC-ILC tracers for {config.experiment} experiment and {''.join(config.foreground_models)} foreground model", verbose=config_mc.verbose)

    #mc_data.nuisance = mc_data.noise + mc_data.cmb

    tracers = component_separation(config_mc, mc_data)
    
    tracers = _combine_B_tracers(np.array(tracers.total))

    _log(f"Saving the tracers in {config.real_mc_tracers[0]['path_tracers']} directory", verbose=config_mc.verbose)

    _save_real_tracers_B(
        tracers,
        config.real_mc_tracers[0]["path_tracers"],
        np.array(config_mc.instrument.channels_tags)[config.real_mc_tracers[0]["channels_tracers"]],
        config_mc.fwhm_out,
        config_mc.lmax
    )

def _save_real_tracers_B(tracers, path_tracers, tags, fwhm_out, lmax):
    """
    Save the generated realistic B-mode tracers in the specified path.

    Parameters
    ----------
        tracers : np.ndarray
            The generated B-mode tracers to be saved. It should be a 2D with shape (n_channels, n_alms).
        path_tracers : str
            The path where the tracers will be saved. 
        tags : List[str]
            List of tags for the tracers, corresponding to the frequency channels of the tracers.
        fwhm_out : float
            Full width at half maximum of the output beam in arcminutes associated to the tracers.
        lmax : int
            Maximum multipole for the tracers.
    
    Returns
    -------
        None
            Saves the tracers in the specified path with the format "B_tracer_{tag}_{fwhm_out}acm_ns{nside}_lmax{lmax}.fits"

    """

    if not os.path.exists(path_tracers):
        os.makedirs(path_tracers)

    if not path_tracers.endswith('/'):
        path_tracers = path_tracers + '/'

    for i, tracer in enumerate(tracers):
        hp.write_map(path_tracers + f"B_tracer_{tags[i]}_{fwhm_out}acm_ns{hp.npix2nside(tracer.shape[0])}_lmax{lmax}.fits", tracer, overwrite=True)
    
def initialize_scalar_tracers(
    config: Configs,
    input_alms: np.ndarray,
    compsep_run: Dict[str, Any],
    field: str = "B",
    **kwargs: Any
) -> np.ndarray:
    """
    Return the scalar tracers for the MC-ILC component separation run.

    Parameters
    ----------
        config : Configs
            Configuration object containing the instrumental and parameters configuration. It should contain the following attributes:
            - `nside`: HEALPix resolution parameter.
            - `lmax`: Maximum multipole for the tracers.
        input_alms : np.ndarray
            Input foreground alms from which the tracers will be derived if `compsep_run["mc_type"]` is "cea_ideal" or "rp_ideal".
            It should have shape (n_channels, n_alms).
        compsep_run : dict
            Dictionary containing the information about the component separation run. It should contain the following keys:
            - `mc_type`: Type of MC-ILC tracer and partition, either "cea_ideal", "rp_ideal", "cea_real", or "rp_real".
            - `domain`: Domain where component separation is performed, either "needlet" or "pixel".
            - `channels_tracers`: List of channels indexes to be used for the tracers generation or loading.
            - `path_tracers`: Path where the tracers will be loaded from if `compsep_run["mc_type"]` is "cea_real" or "rp_real".
        field : str, optional
            Field of the tracers to be loaded or derived, only "B" is supported. Default is "B".
        kwargs : dict, optional
            Additional keyword arguments to be passed to healpy 'map2alm' function.
    
    Returns
    -------
        tracers : np.ndarray
            The scalar tracers for the MC-ILC component separation run. 
            If compsep_run["domain"] is "needlet", it will be a 2D array with alms of the tracers.
            If compsep_run["domain"] is "pixel", it will be a 2D array with maps of the tracers.
    """

    if compsep_run["mc_type"] in ["cea_ideal", "rp_ideal"]:
        if compsep_run["domain"] == "needlet":
            tracers = np.array([
                input_alms[compsep_run["channels_tracers"][0], :],
                input_alms[compsep_run["channels_tracers"][1], :]
            ])
        elif compsep_run["domain"] == "pixel":
            tracers = np.array([
                hp.alm2map(input_alms[compsep_run["channels_tracers"][0], :], config.nside, lmax=config.lmax, pol=False),
                hp.alm2map(input_alms[compsep_run["channels_tracers"][1], :], config.nside, lmax=config.lmax, pol=False)
            ])
    if compsep_run["mc_type"] in ["cea_real", "rp_real"]:
        tracers_tags = [config.instrument.channels_tags[freq] for freq in compsep_run["channels_tracers"]]
        tracer_paths = get_tracers_paths_for_ratio(config, compsep_run["path_tracers"], tracers_tags, field=field)
        if compsep_run["domain"] == "needlet":
            tracers = load_scalar_tracers_for_ratio(tracer_paths, config.nside, config.lmax, return_alms=True, **kwargs)
        elif compsep_run["domain"] == "pixel":
            tracers = load_scalar_tracers_for_ratio(tracer_paths, config.nside, config.lmax, return_alms=False, **kwargs)
    return tracers

def get_tracers_paths_for_ratio(
    config: Configs,
    path_tracers: str,
    tracers_tags: List[str],
    field: str = "B"
) -> List[str]:
    """
    Get the paths to the scalar tracers for the MC-ILC component separation run.

    Parameters
    ----------
        config : Configs
            Configuration object containing the instrumental and parameters configuration. It should contain the following attributes:
            - `nside`: HEALPix resolution parameter.
            - `lmax`: Maximum multipole for the tracers.
            - `fwhm_out`: Full width at half maximum of the output beam in arcminutes associated to the tracers.
        path_tracers : str
            Path where the tracers are stored. It should be a directory path.
        tracers_tags : List[str]
            List of tags for the tracers, corresponding to the frequency channels of the tracers.
        field : str, optional
            Field of the tracers to be loaded, only "B" is supported. Default is "B".
    
    Returns
    -------
        tracers_paths : List[str]
            List of paths to the scalar tracers for the MC-ILC component separation run.
    """

    tracers_paths = [
        f"{path_tracers}{field}_tracer_{tag}_{config.fwhm_out}acm_ns{config.nside}_lmax{config.lmax}.fits"
        for tag in tracers_tags
    ]

    missing_tracers = [tracers_tags[n] for n, path in enumerate(tracers_paths) if not os.path.exists(path)]

    if missing_tracers:
        raise ValueError(f"Missing tracer files: {missing_tracers}. Please check the paths or run the tracer generation routine.")

    return tracers_paths

def load_scalar_tracers_for_ratio(
    tracers_paths: List[str],
    nside: int,
    lmax: int,
    field: str = "B",
    return_alms: bool = True,
    **kwargs: Any
) -> np.ndarray:
    """
    Load the scalar tracers from the specified paths.

    Parameters
    ----------
        tracers_paths : List[str]
            List of paths to the scalar tracers to be loaded.
        nside : int
            HEALPix resolution parameter for the tracers.
        lmax : int
            Maximum multipole for the tracers.
        field : str, optional
            Field of the tracers to be loaded, only "B" is supported. Default is "B".
        return_alms : bool, optional
            Whether to return the tracers as alms or maps. If True, returns alms; if False, returns maps. Default is True.
        kwargs : dict, optional
            Additional keyword arguments to be passed to healpy 'map2alm' function.
    
    Returns
    -------
        tracers : np.ndarray
            The loaded scalar tracers. If `return_alms` is True, it will be a 2D array with alms of the tracers.
            If `return_alms` is False, it will be a 2D array with maps of the tracers.
    """

    tracers = []
    for tracer_path in tracers_paths:
        tracer = hp.read_map(tracer_path, field=0)
        if return_alms:
            tracer = hp.map2alm(tracer, lmax=lmax, pol=False, **kwargs)
        else:
            if hp.get_nside(tracer) != nside:
                alm_ = hp.map2alm(tracer, lmax=lmax, pol=False, **kwargs)
                tracer = hp.alm2map(alm_, nside, lmax=lmax, pol=False)
        tracers.append(tracer)

    return np.array(tracers)     

def get_scalar_tracer_nl(tracers, nside_, lmax_, b_ell):
    """
    Get the scalar MC-ILC ratio tracer in the needlet domain.

    Parameters
    ----------
        tracers : np.ndarray
            The scalar tracers to be filtered. It can be either a 1D array (single tracer) or a 2D array (multiple tracers).
            If 2D, the ratio of the first tracer to the second tracer will be returned.
        nside_ : int
            HEALPix resolution parameter for the tracers.
        lmax_ : int
            Maximum multipole for the tracers.
        b_ell : np.ndarray
            Needlet filter bandpass function to be applied to the tracers. 
    
    Returns
    -------
        np.ndarray
            The filtered scalar tracer in the needlet domain. 
    """

    if tracers.ndim == 1:
        alm_tracer_nl = _needlet_filtering(tracers, b_ell, lmax_)
        return hp.alm2map(alm_tracer_nl, nside_, lmax=lmax_, pol=False)
    elif tracers.ndim == 2:
        tracers_nl = []
        for tracer in tracers:
            alm_tracer_nl = _needlet_filtering(tracer, b_ell, lmax_)
            tracers_nl.append(hp.alm2map(alm_tracer_nl, nside_, lmax=lmax_, pol=False))
        tracers_nl = np.array(tracers_nl)
        return tracers_nl[0] / tracers_nl[1]

def get_scalar_tracer(
    tracers: np.ndarray
) -> np.ndarray:
    """
    Get the scalar MC-ILC ratio tracer in the pixel domain.

    Parameters
    ----------
        tracers : np.ndarray
            The scalar tracers. It can be either a 1D array (single tracer) or a 2D array (multiple tracers).
            If 2D, the ratio of the first tracer to the second tracer will be returned.
    
    Returns
    -------
        np.ndarray
            The scalar tracer in the pixel domain. If `tracers` is a 1D array, it returns the tracer as is.
    """

    if tracers.ndim == 1:
        return tracers
    elif tracers.ndim == 2:
        return tracers[0] / tracers[1]

def _cea_partition(
    map_: np.ndarray,
    n_patches: int,
    mask: Optional[np.ndarray] = None
) -> np.ndarray:
    """
    Partition the input map into n_patches using the CEA method.
    The CEA method splits the map into n_patches patches with equal number of pixels based on the values of the map.

    Parameters
    ----------
        map_ : np.ndarray
            The input map to be partitioned. It should be a 1D array of pixel values.
        n_patches : int
            The number of patches to be created from the input map.
        mask : np.ndarray, optional
            A mask to be applied to the input map. If provided, only the pixels where the mask is non-zero will be considered for partitioning.
            If None, the entire map will be used for partitioning.
    
    Returns
    -------
        np.ndarray
            A 1D array of the same shape as the input map, where each pixel is assigned a patch number from 0 to n_patches-1.
    """
    if mask is None:
        mask = np.ones_like(map_)

    split = np.array_split(np.sort(map_[mask > 0.]),n_patches)
    patches = np.zeros(12 * (hp.get_nside(map_))**2)
    
    for n in range(n_patches):
        if n==0:
            patches[(map_ <= max(split[n])) & (mask > 0.)] = float(n)
        elif n==(n_patches-1):
            patches[(map_ >= min(split[n])) & (mask > 0.)] = float(n)
        else:
            patches[(min(split[n]) <= map_) & (map_ <= max(split[n])) & (mask > 0.)] = float(n)
            
    return patches

def _rp_partition(map_: np.ndarray,
    n_patches: int,
    mask: Optional[np.ndarray] = None) -> np.ndarray:
    """
    Partition the input map into n_patches using the RP method.
    The RP method splits the map into n_patches patches with random number of pixels based on the values of the map.

    Parameters
    ----------
        map_ : np.ndarray
            The input map to be partitioned. It should be a 1D array of pixel values.
        n_patches : int
            The number of patches to be created from the input map.
        mask : np.ndarray, optional
            A mask to be applied to the input map. If provided, only the pixels where the mask is non-zero will be considered for partitioning.
    
    Returns
    -------
        np.ndarray
            A 1D array of the same shape as the input map, where each pixel is assigned a patch number from 0 to n_patches-1.
    """
    if mask is None:
        mask = np.ones_like(map_)

    min_fraction = (0.05 * (np.sum(mask > 0.)/mask.shape[0]) )/ n_patches
    while True:
        partition = np.random.uniform(low=0.0, high=1.0, size=n_patches)
        partition = partition / np.sum(partition)
        if np.all(partition >= min_fraction):
            break

    partition = np.cumsum(partition)
    
    # Define index bins for partitioning the sorted map
#    total_pixels = map_.shape[0]
#    sorted_indices = np.argsort(map_)
#    bins = [0] + [int(total_pixels * partition[i]) for i in range(n_patches - 1)] + [total_pixels]

#    patches = np.zeros(12 * (hp.get_nside(map_))**2)
#    for i in range(n_patches):
#        patches[sorted_indices[bins[i]:bins[i+1]]]=i

    unmasked_indices = np.where(mask > 0.)[0]
    unmasked_map = map_[unmasked_indices]

    # Sort only the unmasked values
    sorted_unmasked_indices = unmasked_indices[np.argsort(unmasked_map)]

    total_unmasked = len(unmasked_indices)
    bins = [0] + [int(total_unmasked * partition[i]) for i in range(n_patches - 1)] + [total_unmasked]

    patches = np.zeros(12 * (hp.get_nside(map_))**2)

    # Assign patch indices only to unmasked pixels
    for i in range(n_patches):
        patch_indices = sorted_unmasked_indices[bins[i]:bins[i+1]]
        patches[patch_indices] = i

    return patches

def _adapt_tracers_path(
    path_tracers: Union[str, List[str]],
    n_fields: int = 1
) -> Union[str, List[str]]:
    """
    Adapt the path_tracers to the number of fields to be analysed.
    If n_fields is 1, it returns a single path as a string.
    If n_fields is greater than 1, it returns a list of paths with the same length as n_fields.

    Parameters
    ----------
        path_tracers : Union[str, List[str]]
            The path or list of paths to the tracers. It can be a single string or a list of strings.
        n_fields : int, optional
            The number of fields to be analysed. Default is 1.
        
    Returns
    -------
        Union[str, List[str]]
            The adapted path or list of paths to the tracers. If n_fields is 1, it returns a single string.
            If n_fields is greater than 1, it returns a list of strings with the same length as n_fields.
    """

    if n_fields == 1:
        if isinstance(path_tracers, str):
            return path_tracers
        elif isinstance(path_tracers, list):
            if len(path_tracers) == 1:
                return path_tracers[0]
            else:
                raise ValueError("If path_tracers is a list, it must contain only one element if you want to analyse just one field.")
    else:
        if isinstance(path_tracers, str):
            return [path_tracers] * n_fields
        elif isinstance(path_tracers, list):
            if len(path_tracers) == n_fields:
                return path_tracers
            else:
                raise ValueError("If path_tracers is a list, it must contain as many elements as the number of fields to be analysed.")

def get_tracers_compsep(
    channels_tracers: List[int],
    lmax: int
) -> List[Dict[str, Any]]:
    """
    Configure component separation methods for obtaining scalar tracers using GILC and needlet domains.

    Parameters
    ----------
        channels_tracers : List[int]
            List of channels indexes associated to the foreground tracers to be used in the component separation.
        lmax : int
            Maximum multipole for the component separation and analysis.
    
    Returns
    -------
        List[Dict[str, Any]]
            List of dictionaries containing the configuration for GILC run to obtain the scalar tracers.
    """    
    def generate_merging_needlets(starting_list: List[int]) -> List[int]:
        needlet_config = {
            "needlet_windows": "mexican",
            "width": 1.3,
            "merging_needlets": starting_list.copy(),
        }
        bl = _get_needlet_windows_(needlet_config, lmax)
        while np.abs(np.sum(bl ** 2, axis=0)[-1] - 1.0) > 1e-5:
            needlet_config["merging_needlets"].append(needlet_config["merging_needlets"][-1] + 5)
            bl = _get_needlet_windows_(needlet_config, lmax)
        return needlet_config["merging_needlets"]

    merging_needlets_1 = generate_merging_needlets([0, 16, 20, 23, 25, 27, 29, 40])
    merging_needlets_2 = generate_merging_needlets([0, 12, 15, 20, 23, 25, 27, 29, 40])


    tracers_compsep = [
    {"method": "gilc",
    "domain": "needlet",
    "ilc_bias": 0.13,
    "needlet_config":[
    {"needlet_windows": "mexican",
    "width": 1.3,
    "merging_needlets": merging_needlets_1}],
    "channels_out": channels_tracers,
    },
    {"method": "gilc",
    "domain": "needlet",
    "ilc_bias": 0.13,
    "needlet_config":[
    {"needlet_windows": "mexican",
    "width": 1.3,
    "merging_needlets": merging_needlets_2}],
    "channels_out": channels_tracers,
    }]
    return tracers_compsep

def get_mc_config(config: Configs, tracers_inputs_path: str) -> Configs:
    """
    Generate a configuration object for the MC-ILC tracers generation based on the provided configuration.

    Parameters
    ----------
    config : Configs
            Configuration object containing the instrumental and parameters configuration. 
            See 'generate_and_save_real_tracers_B' for details.
        tracers_inputs_path : str
            Path where the MC-ILC tracers inputs are stored. It should be a directory path.

    Returns
    -------
        Configs
            A new configuration object for the MC-ILC tracers generation, with paths and parameters set according to the provided configuration.
    """

    config_mc = Configs(config=config.to_dict_for_mc())
    config_mc.return_fgd_components = False

#    if config_mc.bandpass_integrate:
#        config_mc.data_path = f"inputs_mc_tracers/{config_mc.experiment}/total/{''.join(config_mc.foreground_models)}/total_bp_{config_mc.data_type}_ns{config_mc.nside}_lmax{config_mc.lmax}"
#        config_mc.fgds_path = f"inputs_mc_tracers/{config_mc.experiment}/foregrounds/{''.join(config_mc.foreground_models)}/foregrounds_bp_{config_mc.data_type}_ns{config_mc.nside}_lmax{config_mc.lmax}"
#    else:
#        config_mc.data_path = f"inputs_mc_tracers/{config_mc.experiment}/total/{''.join(config_mc.foreground_models)}/total_{config_mc.data_type}_ns{config_mc.nside}_lmax{config_mc.lmax}"
#        config_mc.fgds_path = f"inputs_mc_tracers/{config_mc.experiment}/foregrounds/{''.join(config_mc.foreground_models)}/foregrounds_{config_mc.data_type}_ns{config_mc.nside}_lmax{config_mc.lmax}"
#    config_mc.noise_path = f"inputs_mc_tracers/{config_mc.experiment}/noise/noise_{config_mc.data_type}_ns{config_mc.nside}_lmax{config_mc.lmax}"
#    config_mc.cmb_path = f"inputs_mc_tracers/{config_mc.experiment}/cmb/cmb_{config_mc.data_type}_ns{config_mc.nside}_lmax{config_mc.lmax}"
    if config_mc.bandpass_integrate:
        config_mc.data_path = f"{tracers_inputs_path}/total_{''.join(config_mc.foreground_models)}_bp_{config_mc.data_type}_ns{config_mc.nside}_lmax{config_mc.lmax}"
        config_mc.fgds_path = f"{tracers_inputs_path}/foregrounds_{''.join(config_mc.foreground_models)}_bp_{config_mc.data_type}_ns{config_mc.nside}_lmax{config_mc.lmax}"
    else:
        config_mc.data_path = f"{tracers_inputs_path}/total_{''.join(config_mc.foreground_models)}_{config_mc.data_type}_ns{config_mc.nside}_lmax{config_mc.lmax}"
        config_mc.fgds_path = f"{tracers_inputs_path}/foregrounds_{''.join(config_mc.foreground_models)}_{config_mc.data_type}_ns{config_mc.nside}_lmax{config_mc.lmax}"
    config_mc.noise_path = f"{tracers_inputs_path}/noise_{config_mc.data_type}_ns{config_mc.nside}_lmax{config_mc.lmax}"
    config_mc.cmb_path = f"{tracers_inputs_path}/cmb_{config_mc.data_type}_ns{config_mc.nside}_lmax{config_mc.lmax}"
    
#    if not all(os.path.exists(p + ".npy") for p in [config_mc.data_path, config_mc.noise_path, config_mc.cmb_path]):
#        if not os.path.exists(config_mc.fgds_path + f"_{''.join(config_mc.foreground_models)}.npy"):
    config_mc.generate_input_foregrounds = True #
#        else:
#            config_mc.generate_input_foregrounds = False
    config_mc.generate_input_noise = True #
    config_mc.generate_input_cmb = True #
    config_mc.generate_input_data = True #
    config_mc.save_inputs = True # 
    config_mc.seed_cmb = None
    config_mc.seed_noise = None
#    else:
#        if not os.path.exists(config_mc.fgds_path + f"_{''.join(config_mc.foreground_models)}.npy"):
#            config_mc.generate_input_foregrounds = True
#            config_mc.generate_input_data = True #
#            config_mc.save_inputs = True
#        else:
#            config_mc.generate_input_foregrounds = False
#            config_mc.generate_input_data = False
#            config_mc.save_inputs = False #
#       config_mc.generate_input_noise = False #
#       config_mc.generate_input_cmb = False #

    config_mc.save_compsep_products = False #
    config_mc.return_compsep_products = True
#    config_mc.mask_type = "mask_for_compsep"

    if config_mc.data_type == "maps":
        config_mc.field_in = "QU"
        config_mc.mc_data_field = "TQU"
    elif config_mc.data_type == "alms":
        config_mc.field_in = "B"
        config_mc.mc_data_field = "TEB"
    config_mc.field_out = "B"

    return config_mc

def get_mc_data(
    config_mc: "Configs",
    foregrounds: Optional[np.ndarray] = None,
    systematics: Optional[np.ndarray] = None,
    **kwargs
) -> Any:
    """
    Simulate the input data for the MC-ILC tracers generation based on the provided configuration.

    Parameters
    ----------
        config_mc : Configs
            Configuration object for the MC-ILC tracers generation. See 'get_mc_config' for details.
        foregrounds : SimpleNamespace, optional
            Foregrounds data to be used in the simulations for tracer generation. If None, the foregrounds will be generated using the configuration parameters.
        kwargs : dict, optional
            Additional keyword arguments to be passed to healpy 'map2alm' function.
    
    Returns
    -------
        SimpleNamespace
            A SimpleNamespace object containing the simulated data for the MC-ILC tracers generation. 
            It will have the following attributes:
            - `total`: The total simulated data, which includes CMB, foregrounds, and noise.
            - `cmb`: The simulated CMB data.
            - `fgds`: The simulated foregrounds data.
            - `noise`: The simulated noise data.
    """

    if foregrounds is None:
        mc_foregrounds = _get_data_foregrounds_(config_mc)
    else:
        mc_foregrounds = SimpleNamespace()
        mc_foregrounds.total = foregrounds

    mc_data = _get_data_simulations_(config_mc, mc_foregrounds)

    if systematics is not None:
        mc_data.total = mc_data.total + systematics

    return _slice_data(mc_data, config_mc.mc_data_field, config_mc.field_in)

def _combine_B_tracers(tracers, coefficients=[0.7,0.3]):
    """
    Combine the scalar tracers using the provided coefficients.

    Parameters
    ----------
        tracers : np.ndarray
            The scalar tracers to be combined. It can be either a 3D array (multiple tracers) or a 2D array (single tracer).
            If 2D, it should have shape (n_channels, n_pixels).
        coefficients : List[float], optional
            List of coefficients to be used for combining the tracers. Default is [0.7, 0.3].
            The coefficients should sum to 1. If they do not, they will be normalized.
    
    Returns
    -------
        np.ndarray
            The combined scalar tracers. If `tracers` is a 2D array, it returns the tracers as is.
            If `tracers` is a 3D array, it returns the weighted sum of the tracers using the provided coefficients. 
    """

    if tracers.ndim == 2:
        return tracers
    elif tracers.ndim == 3:
        if len(coefficients) != tracers.shape[0]:
            raise ValueError("The number of coefficients must match the number of tracers.")
        coefficients = np.array(coefficients) / np.sum(coefficients)
        return np.einsum("i,ijk->jk", coefficients, tracers)

__all__ = [
    name
    for name, obj in globals().items()
    if callable(obj) and getattr(obj, "__module__", None) == __name__
]



import os
import numpy as np
import healpy as hp
import pysm3
import pysm3.units as u
from astropy.io import fits
from types import SimpleNamespace
from typing import Optional, Union, List, Dict, Any
import warnings
from threadpoolctl import threadpool_limits
import sys
from .configurations import Configs
from .seds import _get_CMB_SED
from .routines import (
    _get_ell_filter,
    _get_beam_from_file,
    _map2alm_kwargs, _log, _get_bandwidths,
)

prefix_to_attr = {
                "d": "dust", "s": "synch", "a": "ame", "co": "co",
                "f": "freefree", "cib": "cib", "tsz": "tsz",
                "ksz": "ksz", "rg": "radio_galaxies"
            }

def _get_full_simulations(config: Configs, nsim: Optional[Union[int, str]] = None, **kwargs: Any) -> SimpleNamespace:
    """
    Generate full simulations including foregrounds and data.

    Parameters
    ----------
        config: Configs
            Configuration parameters.
        nsim: Optional[Union[int, str]]
            Simulation number.
        kwargs: dict, optional
            Additional keyword arguments forwarded to alm computation in map2alm.

    Returns
    -------
        SimpleNamespace
            Simulated data container with foregrounds and total data.
    """
    kwargs = _map2alm_kwargs(**kwargs)

    foregrounds = _get_data_foregrounds_(config, **kwargs)

    data = _get_data_simulations_(config, foregrounds, nsim=nsim, **kwargs)

    return data

def _get_data_foregrounds_(config: Configs, **kwargs: Any) -> SimpleNamespace:
    """
    Load or generate foreground maps based on configuration.

    Parameters
    ----------
        config: Configs
            Configuration parameters. It should have the following attributes:
            - `generate_input_foregrounds`: Whether to generate foreground maps.
            - `foreground_models`: List of foreground models to generate.
            - 'instrument': a dictionary containing the instrument configuration, including:
                - `frequency`: List of instrument frequencies in GHz.
                - `beams`: Type of beams to be used (e.g., "gaussian", "file_l", "file_lm").
                - `fwhm`: List of full width at half maximum (FWHM) for each frequency channel in arcmin. Used if beams are "gaussian".
                - 'depth_I': Depth for intensity maps in arcmin*uK_CMB (optional). 
                            If not provided, it will be assumed to be the polarization depth divided by sqrt(2).
                            Used if path_depth_maps is not provided.
                - 'depth_P': Depth for polarization maps in arcmin*uK_CMB (optional).
                            If not provided, it will be assumed to be the intensity depth multiplied by sqrt(2).
                            Used if path_depth_maps is not provided.
                - `path_beams`: Full path to the beams files (if using "file_l" or "file_lm" beams). 
                            The code will look for files named "{path_beams}_{channel_tag}.fits" for each frequency channel.
                - `channels_tags`: List of tags for each frequency channel, used for loading beams, bandpasses or depth maps.
                - 'bandwidths': List of relative bandwidths for each frequency channel (optional, used if bandpass_integrate is True).
                            Used if path_bandpasses is not provided.
                - `path_depth_maps`: Full path to depth maps (optional, used if generating noise). 
                            The code will look for files named "{path_depth_maps}_{channel_tag}.fits" for each frequency channel.
                - `path_hits_maps`: Full path to hits maps (optional, used if generating noise and 'depth_maps' is not provided).
                            If it does not end with .fits, the code will look for files named "{path_hits_maps}_{channel_tag}.fits" for each frequency channel.
                - `path_bandpasses`: Full path to bandpass files (optional, used if bandpass_integrate is True).
                            It will look for files named as "{path_bandpasses}_{channel_tag}.npy" for each channel tag.
                            Each file should be a 2D array which has the first column a list of frequencies in GHz and the second column the corresponding bandpass response.
                - `ell_knee`: Lists of knee frequencies for each channel for the noise power spectrum (optional).
                            If it is a single list it will be applied to temperature only.
                            If it is a list of two lists it will be applied to temperature (first list) and polarization (second list).
                            If not provided, white noise is assumed.
                - `alpha_knee`: List of spectral indices of the noise power spectrum for each channel (optional).
                            If not provided, white noise is assumed.
            - `nside`: HEALPix resolution.
            - `lmax`: Maximum multipole for the simulation.
            - `return_fgd_components`: Whether to return individual foreground components.
            - `fgds_path`: Path where saving or loading foreground maps.
            - `save_inputs`: Whether to save generated foreground maps to disk.        
            - `pixel_window_in`: Whether to apply pixel window smoothing.
            - `units`: Units for the foreground maps (e.g., 'uK_CMB').
            - `data_type`: Type of data to return, either "maps" or "alms".
            - `bandpass_integrate`: Whether to integrate foreground components across bandpasses.
            - `lmin`: Minimum multipole to keep in the simulation.
            - `coordinates`: Coordinate system for the maps (e.g., "G" for Galactic).
        kwargs: dict, optional
            Additional keyword arguments forwarded to alm computation.

    Returns
    -------
        SimpleNamespace
            Foregrounds object containing single components (optionally) and total map.
    """
    kwargs = _map2alm_kwargs(**kwargs)

    if config.generate_input_foregrounds:
        if config.verbose:
            msg = f"Generating foreground maps of {''.join(config.foreground_models)} model"
            if config.bandpass_integrate:
                msg += " with bandpass integration"
            print(msg)

        foregrounds = _get_foregrounds(
            config.foreground_models,
            config.instrument,
            config.nside,
            config.lmax,
            return_components=config.return_fgd_components,
            pixel_window=config.pixel_window_in,
            units=config.units,
            return_alms=(config.data_type == "alms"),
            bandpass_integrate=config.bandpass_integrate,
            lmin=config.lmin,
            coordinates=config.coordinates,
            **kwargs
        )
        
        if config.save_inputs:
            _log(f"Saving foreground maps in {config.fgds_path} directory", verbose=config.verbose)
            _save_input_foregrounds(config.fgds_path, foregrounds, config.foreground_models)
    else:
        foregrounds = SimpleNamespace()
        if config.return_fgd_components:
            for fmodel in config.foreground_models:
                attr = prefix_to_attr.get(fmodel[:3]) or prefix_to_attr.get(fmodel[:2]) or prefix_to_attr.get(fmodel[:1])
                setattr(foregrounds, attr, _load_input_foregrounds(config.fgds_path, fmodel))
        foregrounds.total = _load_input_foregrounds(config.fgds_path, "".join(config.foreground_models))
    return foregrounds


def _get_data_simulations_(
    config: Configs,
    foregrounds: Optional[SimpleNamespace] = None,
    nsim: Optional[Union[int, str]] = None,
    **kwargs: Any
    ) -> SimpleNamespace:
    """
    Load or generate simulation data including CMB, noise, and combined total.

    Parameters
    ----------
        config: Configs
            Configuration parameters. It should have the following attributes:
            - `generate_input_cmb`: Whether to generate CMB maps. If False, it will load from `cmb_path`.
            - `cmb_path`: Path where saving or loading CMB maps.
            - 'cls_cmb_path': Path to the CMB power spectrum FITS file. Used if 'generate_input_cmb' is True.
            - 'seed_cmb': Seed for CMB generation (optional).
            - 'cls_cmb_new_ordered': Whether the new ordering of Cls is used in the CMB power spectrum FITS file.
            - `generate_input_noise`: Whether to generate noise maps. If False, it will load from `noise_path`.
            - `noise_path`: Path where saving or loading noise maps.
            - `seed_noise`: Seed for noise generation (optional).
            - `generate_input_data`: Whether to generate total data maps. If False, it will load from `data_path`.
            - `data_path`: Path where saving or loading total data maps.
            - `save_inputs`: Whether to save generated inputs to disk.
            - `lmax`: Maximum multipole for the simulation.
            - `nside`: Desired HEALPix resolution.
            - `data_type`: Type of data to return, either "maps" or "alms". It must be compatible with provided foregrounds, if any.
            - `units`: Units for the maps (e.g., 'uK_CMB').
            - `lmin`: Minimum multipole to keep in the simulation. Default is 2.
            - `pixel_window_in`: Whether to apply pixel window smoothing to the input maps.
            - 'instrument': a dictionary containing the instrument configuration, including:
                - `frequency`: List of instrument frequencies in GHz.
                - `beams`: Type of beams to be used (e.g., "gaussian", "file_l", "file_lm").
                - `fwhm`: List of full width at half maximum (FWHM) for each frequency channel in arcmin. Used if beams are "gaussian".
                - 'depth_I': Depth for intensity maps in arcmin*uK_CMB (optional). 
                            If not provided, it will be assumed to be the polarization depth divided by sqrt(2).
                            Used if path_depth_maps is not provided.
                - 'depth_P': Depth for polarization maps in arcmin*uK_CMB (optional).
                            If not provided, it will be assumed to be the intensity depth multiplied by sqrt(2).
                            Used if path_depth_maps is not provided.
                - `path_beams`: Path to the beam files (if using "file_l" or "file_lm" beams).
                            The code will look for files named "{path_beams}_{channel_tag}.fits" for each frequency channel.
                - `channels_tags`: List of tags for each frequency channel, used for loading beams, bandpasses or depth maps.
                - 'bandwidths': List of relative bandwidths for each frequency channel (optional, used if bandpass_integrate is True).
                            Used if path_bandpasses is not provided.
                - `path_depth_maps`: Full path to depth maps (optional, used if generating noise).
                            The code will look for files named "{path_depth_maps}_{channel_tag}.fits" for each frequency channel.   
                - `path_hits_maps`: Full path to hits maps (optional, used if generating noise and 'path_depth_maps' is not provided).
                            If it does not end with .fits, the code will look for files named 
                            "{path_hits_maps}_{channel_tag}.fits" for each frequency channel.
                - `path_bandpasses`: Path to bandpass files (optional, used if bandpass_integrate is True).
                            The code will look for files named as "{path_bandpasses}_{channel_tag}.npy" for each channel tag.
                            Each file should be a 2D array which has the first column a list of frequencies in GHz and the second column the corresponding bandpass response.
                - `ell_knee`: Lists of knee frequencies for each channel for the noise power spectrum (optional).
                            If it is a single list it will be applied to temperature only.
                            If it is a list of two lists it will be applied to temperature (first list) and polarization (second list).
                            If not provided, white noise is assumed.
                - `alpha_knee`: List of spectral indices of the noise power spectrum for each channel (optional).
                            If not provided, white noise is assumed.
        foregrounds: Optional[SimpleNamespace]
            Foreground components.
        nsim: Optional[Union[int, str]]
            Simulation number.
        kwargs: dict, optional
            Additional keyword arguments forwarded to alm computation.

    Returns
    -------
        SimpleNamespace
            Data container with cmb, noise, total and foregrounds.
    """
    if nsim is not None:
        if not isinstance(nsim, (int, str)):
            raise ValueError("nsim must be an integer or a string.")
        if isinstance(nsim, int):
            nsim = str(nsim).zfill(5)
    
    kwargs = _map2alm_kwargs(**kwargs)

    if foregrounds is not None and not hasattr(foregrounds, 'total'):
        raise ValueError('foregrounds must have the attribute total.')

    data = SimpleNamespace()

    if config.generate_input_cmb:
        _log(f"Generating CMB simulation" + f" {nsim}" if nsim is not None else "", verbose=config.verbose)
        data.cmb = _get_cmb_simulation(config, nsim=nsim)
    elif config.cmb_path is not None:
        if config.verbose:
            path_str = f"{config.cmb_path}.npy" if nsim is None else f"{config.cmb_path}_{nsim}.npy"
            print(f"Loading CMB from {path_str}")
        data.cmb = _load_inputs(config.cmb_path, nsim=nsim)

    if config.generate_input_noise:
        _log(f"Generating noise simulation" + f" {nsim}" if nsim is not None else "", verbose=config.verbose)
        data.noise = _get_noise_simulation(config, nsim=nsim, **kwargs)
    elif config.noise_path is not None:
        if config.verbose:
            path_str = f"{config.noise_path}.npy" if nsim is None else f"{config.noise_path}_{nsim}.npy"
            print(f"Loading noise from {path_str}")
        data.noise = _load_inputs(config.noise_path, nsim=nsim)

    if config.generate_input_data:
        _log(f"Generating coadded signal" + f" for simulation {nsim}" if nsim is not None else "", verbose=config.verbose)
        if hasattr(data, 'cmb') and hasattr(data, 'noise') and foregrounds is not None:
            data.total = data.noise + data.cmb + foregrounds.total
            if config.save_inputs:
                _save_inputs(config.data_path, data.total, nsim=nsim)
        else:
            raise ValueError("To generate input data, provide foregrounds and CMB/noise paths or generate them.")
    else:
        data.total = _load_inputs(config.data_path, nsim=nsim)

    if foregrounds is not None:
        data.fgds = foregrounds.total

    return data

def _save_inputs(filename: str, maps: np.ndarray, nsim: Union[str, None] = None) -> None:
    """Save simulation maps to disk, creating directories if needed.
    
    Parameters
    ----------
        filename: str
            Path to save the simulation maps, without extension.
        maps: np.ndarray
            Simulation maps to save. Shape is (n_freq, 3, n_pix) for maps or (n_freq, 3, n_alm) for alms.
        nsim: Union[str, None], optional
            Simulation index to append to the filename (optional). Default is None.
        
    Returns
    -------
        None
    
    """
    os.makedirs(os.path.dirname(filename), exist_ok=True)
    if nsim is not None:
        filename += f"_{nsim}"
    np.save(filename, maps)

def _load_inputs(path: str, nsim: Union[str, None] = None) -> np.ndarray:
    """
    Load simulation maps from disk, handling nsim suffix.
    
    Parameters
    ----------
        path: str
            Path to the simulation maps file, without extension.
        nsim: Union[str, None], optional
            Simulation index to append to the filename (optional). Default is None.
    
    Returns
    -------
        np.ndarray
            Loaded simulation maps. Shape is (n_freq, 3, n_pix) for maps or (n_freq, 3, n_alm) for alms.
    """

    filepath = path + f"_{nsim}.npy" if nsim is not None else path + '.npy'
    if not os.path.isfile(filepath):
        raise FileNotFoundError(f"File not found: {filepath}")
    return np.load(filepath)
    
def _save_input_foregrounds(fgds_path: str, foregrounds: SimpleNamespace, foreground_models: List[str]) -> None:
    """
    Save foreground maps for all models and the total.
    Assumes 'foregrounds' is a SimpleNamespace or similar with components as attributes.

    Parameters
    ----------
        fgds_path: str
            Path where saving foreground maps, without extension.
        foregrounds: SimpleNamespace
            Foreground components, should have attributes for each model and a 'total' attribute.
        foreground_models: List[str]
            List of foreground model names to save. Each name should match an attribute in 'foregrounds'.

    Returns
    -------
        None    
    """
    os.makedirs(os.path.dirname(fgds_path), exist_ok=True)
    # Save total foreground map
    np.save(fgds_path + f'_{"".join(foreground_models)}', foregrounds.total)

    if len(vars(foregrounds)) > 1:
        fg_attrs = {k: v for k, v in vars(foregrounds).items() if k != "total"}
        if len(fg_attrs) != len(foreground_models):
            raise ValueError(
                f"Number of foreground components ({len(fg_attrs)}) does not match number of models ({len(foreground_models)}).")

        for fmodel in foreground_models:
            # Try matching longest prefix first (3,2,1)
            attr = None
            for length in (3, 2, 1):
                if len(fmodel) > length:
                    prefix = fmodel[:length]
                    attr = prefix_to_attr.get(prefix)
                    if attr is not None:
                        break
            if attr is None:
                raise ValueError(f"Unknown foreground model prefix for '{fmodel}'")
            if attr not in fg_attrs:
                raise ValueError(f"Foreground attribute '{attr}' missing in foregrounds object")

            np.save(fgds_path + f"_{fmodel}", fg_attrs[attr])

def _load_input_foregrounds(fgd_path: str, fgd_model: str) -> np.ndarray:
    """
    Load foreground map for a given model.
    
    Parameters
    ----------
        fgd_path: str
            Path to the foreground maps, without extension.
        fgd_model: str
            Foreground model name to load. It should match the saved file name suffix.

    Returns
        np.ndarray
            Loaded foreground map. Shape is (n_freq, 3, n_pix) for maps or (n_freq, 3, n_alm) for alms.
    """
    filepath = f'{fgd_path}_{fgd_model}.npy'
    if not os.path.isfile(filepath):
        raise FileNotFoundError(f"Foreground file not found: {filepath}")
    return np.load(filepath)

def _get_noise_simulation(config: Configs, nsim: Optional[Union[int, str]] = None, **kwargs) -> np.ndarray:
    """
    Generate noise simulations for the instrument frequencies.

    Parameters
    ----------
        config: Configs
            Configuration parameters including instrument settings. It should have the following attributes:
            - `instrument.frequency`: List of instrument frequencies.
            - `instrument.depth_I`: Depth for intensity maps.
            - `instrument.depth_P`: Depth for polarization maps.
            - `instrument.path_depth_maps`: Path to depth maps (optional).
            - `instrument.path_hits_maps`: Path to hits maps (optional).
            - `nside`: HEALPix resolution.
            - `lmax`: Maximum multipole for the simulation.
            - `data_type`: Type of data to return, either "maps" or "alms".
            - `units`: Units for the noise maps (e.g., 'uK_CMB').
            - `lmin`: Minimum multipole to keep in the simulation.
            - `seed_noise`: Seed for noise generation (optional).    
        nsim: int or str, optional
            Simulation index to save the maps and vary the random seed (optional). Default: None.
        kwargs: dict, optional
            Additional keyword arguments for `hp.map2alm`.

    Returns
    -------
        noise: np.ndarray
            array of noise maps or alms. Shape is (n_freq, 3, n_pix) for maps or (n_freq, 3, n_alm) for alms.
    """

    if nsim is not None:
        if not isinstance(nsim, (int, str)):
            raise ValueError("nsim must be an integer or a string.")
        if isinstance(nsim, int):
            nsim = str(nsim).zfill(5)


    # Setup seed for reproducibility
    if config.seed_noise is None:
        seed = None
    else:
        if nsim is not None:
            seed = config.seed_noise + int(nsim) * 3 * len(config.instrument.frequency)
        else:
            seed = config.seed_noise    

    if not hasattr(config.instrument, 'path_depth_maps'):
        if not hasattr(config.instrument, 'depth_I') and not hasattr(config.instrument, 'depth_P'):
            raise ValueError('Provided instrumental setting must have either depth_I or depth_P attributes.')
        elif not hasattr(config.instrument, 'depth_I') and hasattr(config.instrument, 'depth_P'):
            config.instrument.depth_I = config.instrument.depth_P / np.sqrt(2)
            print('Warning: No intensity map depth provided. Assuming it to be the polarization one divided by sqrt(2).')
        elif not hasattr(config.instrument, 'depth_P') and hasattr(config.instrument, 'depth_I'):
            config.instrument.depth_P = config.instrument.depth_I * np.sqrt(2)
            print('Warning: No polarization map depth provided. Assuming it to be the intensity one multiplied by sqrt(2).')
        depth_i = config.instrument.depth_I
        depth_p = config.instrument.depth_P

        # Load or set hits map if available
        if hasattr(config.instrument, 'path_hits_maps'):
            if config.instrument.path_hits_maps.endswith(".fits"):
                hits_map = hp.read_map(config.instrument.path_hits_maps, field=0, dtype=np.float64)
                if hp.get_nside(hits_map) != config.nside:
                    hits_map = hp.ud_grade(hits_map, nside_out=config.nside, power=-2)
                hits_map /= np.amax(hits_map)
    else:
        depth_i = [1.] * len(config.instrument.frequency)
        depth_p = [1.] * len(config.instrument.frequency)
        
    # Convert depths to requested units with CMB equivalencies
    #depth_i *= u.arcmin * u.uK_CMB
    #depth_i = depth_i.to(getattr(u, config.units) * u.arcmin, equivalencies=u.cmb_equivalencies(config.instrument.frequency * u.GHz))
    #depth_p *= u.arcmin * u.uK_CMB
    #depth_p = depth_p.to(getattr(u, config.units) * u.arcmin, equivalencies=u.cmb_equivalencies(config.instrument.frequency * u.GHz))
    bandwidths = _get_bandwidths(config, np.arange(len(config.instrument.frequency)))
    A_cmb = _get_CMB_SED(config.instrument.frequency, units=config.units, bandwidths=bandwidths)
    depth_i = depth_i * A_cmb
    depth_p = depth_p * A_cmb

    # Precompute conversion factor from arcmin to radians
    acm_to_rad = (np.pi / (180 * 60)) 

    # Get ell filter if needed
    fell = _get_ell_filter(config.lmin, config.lmax) if config.lmin > 2 else None

    noise = []
    for nf, _ in enumerate(config.instrument.frequency):
        # Load depth maps if path provided
        if hasattr(config.instrument, 'path_depth_maps'):
            depth_map_fn = config.instrument.path_depth_maps + f"_{config.instrument.channels_tags[nf]}.fits"
            try:
                depth_maps_in = hp.read_map(depth_map_fn, field=(0,1), dtype=np.float64)
            except IndexError:
                print("Warning: Unable to read depth maps from the provided path for I and P, provided depth map is assumed to refer to polarization.")
                depth_maps_in = hp.read_map(depth_map_fn, field=0, dtype=np.float64)
                depth_maps_in = np.array([depth_maps_in / np.sqrt(2), depth_maps_in])
            if hp.get_nside(depth_maps_in[0]) != config.nside:
                depth_maps = np.array(
                    [np.sqrt(hp.ud_grade(dm**2, nside_out=config.nside, power=2)) for dm in depth_maps_in])
            else:
                depth_maps = np.copy(depth_maps_in)
            del depth_maps_in
        elif hasattr(config.instrument, 'path_hits_maps'):
            if not config.instrument.path_hits_maps.endswith(".fits"):
                hits_file = config.instrument.path_hits_maps + f"_{config.instrument.channels_tags[nf]}.fits"
                hits_map = hp.read_map(hits_file, field=0, dtype=np.float64)
                if hp.get_nside(hits_map) != config.nside:
                    hits_map = hp.ud_grade(hits_map, nside_out=config.nside, power=-2)
                hits_map /= np.amax(hits_map)

        if seed is not None:
            np.random.seed(seed + (nf * 3))
        # Generate noise power spectra
#        N_ell_T = (depth_i.value[nf] * acm_to_rad) ** 2 * np.ones(config.lmax + 1)
#        N_ell_P = (depth_p.value[nf] * acm_to_rad) ** 2 * np.ones(config.lmax + 1)
        N_ell_T = (depth_i[nf] * acm_to_rad) ** 2 * np.ones(config.lmax + 1)
        N_ell_P = (depth_p[nf] * acm_to_rad) ** 2 * np.ones(config.lmax + 1)
        N_ell = np.array([N_ell_T, N_ell_P, N_ell_P, 0.*N_ell_P])

        # Add knee frequency noise if provided
        if hasattr(config.instrument, 'ell_knee') and hasattr(config.instrument, 'alpha_knee'):
            ell = np.arange(config.lmax + 1)
            if isinstance(config.instrument.alpha_knee, list) and isinstance(config.instrument.ell_knee, list):
                if np.array(config.instrument.alpha_knee).ndim == 2 and np.array(config.instrument.ell_knee).ndim == 2:
                    if len(config.instrument.alpha_knee[0]) != len(config.instrument.ell_knee[0]) or len(config.instrument.alpha_knee[1]) != len(config.instrument.ell_knee[1]):
                        raise ValueError('alpha_knee and ell_knee must have the same length.')
                    if (len(config.instrument.ell_knee[0]) != len(config.instrument.frequency)) or (len(config.instrument.ell_knee[1]) != len(config.instrument.frequency)):
                        raise ValueError('alpha_knee and ell_knee must have the same length as the number of frequencies.')
                    N_ell[0] *= (1 + (ell / config.instrument.ell_knee[0][nf]) ** config.instrument.alpha_knee[0][nf])
                    N_ell[1:] *= (1 + (ell / config.instrument.ell_knee[1][nf]) ** config.instrument.alpha_knee[1][nf])
                elif np.array(config.instrument.alpha_knee).ndim == 1 and np.array(config.instrument.ell_knee).ndim == 1:
                    if len(config.instrument.alpha_knee) != len(config.instrument.frequency) or len(config.instrument.ell_knee) != len(config.instrument.frequency):
                        raise ValueError('alpha_knee and ell_knee must have the same length as the number of frequencies.')
                    N_ell[0] *= (1 + (ell / config.instrument.ell_knee[nf]) ** config.instrument.alpha_knee[nf])
                elif np.array(config.instrument.alpha_knee).ndim == 1 and np.array(config.instrument.ell_knee).ndim == 2:
                    if len(config.instrument.alpha_knee) != len(config.instrument.frequency):
                        raise ValueError('alpha_knee must have the same length as the number of frequencies.')
                    if len(config.instrument.ell_knee[0]) != len(config.instrument.frequency) or len(config.instrument.ell_knee[1]) != len(config.instrument.frequency):
                        raise ValueError('ell_knee lists must have the same length as the number of frequencies.')
                    N_ell[0] *= (1 + (ell / config.instrument.ell_knee[0][nf]) ** config.instrument.alpha_knee[nf])
                    N_ell[1:] *= (1 + (ell / config.instrument.ell_knee[1][nf]) ** config.instrument.alpha_knee[nf])
            else:
                raise ValueError('alpha_knee and ell_knee must be both lists or lists of 2 lists')

        # Generate noise alm
        alm_noise = hp.synalm(N_ell, lmax=config.lmax, new=True)

        # Apply ell filter if applicable
        if fell is not None:
            for f in range(3):
                alm_noise[f] = hp.almxfl(alm_noise[f], fell)

        # Generate noise maps or alms depending on data_type
        if config.data_type=="alms":
            if hasattr(config.instrument, 'path_depth_maps'):
                noise_map = hp.alm2map(alm_noise, config.nside, lmax=config.lmax, pol=True)  * np.array([depth_maps[0], depth_maps[1], depth_maps[1]])
                noise.append(hp.map2alm(noise_map, lmax=config.lmax, pol=True, **kwargs))
            elif hasattr(config.instrument, 'path_hits_maps'):
                noise_map = hp.alm2map(alm_noise, config.nside, lmax=config.lmax, pol=True) / np.sqrt(hits_map)
                noise_map[np.isinf(noise_map)] = 0.
                noise.append(hp.map2alm(noise_map, lmax=config.lmax, pol=True, **kwargs))
            else:
                noise.append(alm_noise)
        else:
            if hasattr(config.instrument, 'path_depth_maps'):
                noise.append(hp.alm2map(alm_noise, config.nside, lmax=config.lmax, pol=True) * np.array([depth_maps[0], depth_maps[1], depth_maps[1]]))
            elif hasattr(config.instrument, 'path_hits_maps'):
                noise_map = hp.alm2map(alm_noise, config.nside, lmax=config.lmax, pol=True) / np.sqrt(hits_map)
                noise_map[np.isinf(noise_map)] = 0.
                noise.append(noise_map)
            else:
                noise.append(hp.alm2map(alm_noise, config.nside, lmax=config.lmax, pol=True))

    if config.save_inputs:
        _save_inputs(config.noise_path, np.array(noise), nsim=nsim)

    return np.array(noise)

def _get_cmb_simulation(config: Configs, nsim: Optional[Union[int, str]] = None) -> np.ndarray:
    """
    Generate simulated CMB maps or alms for a given instrument configuration.

    Parameters
    ----------
        config: Configs
            Simulation and instrument configuration. It should have the following attributes:
            - `lmax`: Maximum multipole for the simulation.
            - `nside`: HEALPix resolution.
            - `data_type`: Type of data to return, either "maps" or "alms".
            - `cls_cmb_path`: Path to the CMB power spectrum FITS file.
            - `seed_cmb`: Seed for CMB generation (optional).
            - 'cls_cmb_new_ordered': Whether the new ordering of Cls is used in the CMB power spectrum FITS file.
        nsim: int or str, optional
            Simulation index to save the maps and vary the random seed (optional). Default: None.

    Returns
    -------
        np.ndarray: 
            Simulated CMB maps or harmonic coefficients (alms). Shape is (n_freq, 3, n_pix) for maps or (n_freq, 3, n_alm) for alms.
    """
    # Converting nsim to a string if provided
    if nsim is not None:
        if not isinstance(nsim, (int, str)):
            raise ValueError("nsim must be an integer or a string.")
        if isinstance(nsim, int):
            nsim = str(nsim).zfill(5)

    # Getting default path to CMB power spectrum, if not provided in config
    if not config.cls_cmb_path:
        # Define the path to the FITS file
        raise ValueError(
            "No CMB power spectrum path provided. Please set 'cls_cmb_path' in the configuration."
        )
        
    # Load the CMB power spectrum FITS file (assumed to be in muK_CMB^2 units)
    cls_cmb = hp.read_cl(config.cls_cmb_path)

    # Initializing the seed if required
    seed = None if not config.seed_cmb else (config.seed_cmb + int(nsim) if nsim is not None else config.seed_cmb)
    
    # Generating a realization of CMB alms with the loaded Cls
    alm_cmb = _get_cmb_alms_realization(cls_cmb, config.lmax, seed = seed, new = config.cls_cmb_new_ordered)
    
    # Computing the high-pass filter if lmin > 2
    fell = _get_ell_filter(config.lmin, config.lmax) if config.lmin > 2 else None

    # Smoothing the CMB alms with the beams of each frequency channel
    cmb = []

    bandwidths = _get_bandwidths(config, np.arange(len(config.instrument.frequency)))
    A_cmb = _get_CMB_SED(config.instrument.frequency, units=config.units, bandwidths=bandwidths)
     
    for idx, _ in enumerate(config.instrument.frequency):
        if config.instrument.beams == "gaussian":
            alm_cmb_i = _smooth_input_alms_(
                alm_cmb,
                fwhm=config.instrument.fwhm[idx],
                nside_out=config.nside if config.pixel_window_in else None
            )
        else:
            beamfile = config.instrument.path_beams + f"_{config.instrument.channels_tags[idx]}.fits"
            alm_cmb_i = _smooth_input_alms_(
                alm_cmb,
                beam_path=beamfile,
                symmetric_beam=(config.instrument.beams == "file_l"),
                nside_out=config.nside if config.pixel_window_in else None
            )

        if fell is not None:
            for f in range(3):
                alm_cmb_i[f] = hp.almxfl(alm_cmb_i[f], fell)

        cmb.append(A_cmb[idx] * alm_cmb_i if config.data_type == "alms" else A_cmb[idx] * hp.alm2map(
            alm_cmb_i, config.nside, lmax=config.lmax, pol=True
        ))
    
    cmb = np.array(cmb)

    # Saving the CMB maps/alms if requested
    if config.save_inputs:
        _save_inputs(config.cmb_path, cmb, nsim=nsim)
    return cmb

def _get_cmb_alms_realization(
    cls_cmb: np.ndarray,
    lmax: int,
    seed: Optional[int] = None,
    new: bool = True
) -> np.ndarray:
    """
    Generate a realization of CMB spherical harmonic coefficients (alms).

    Parameters
    ----------
        cls_cmb: np.ndarray
            Theoretical CMB angular power spectra.
        lmax: int
            Maximum multipole for the realization.
        seed: Optional[int]
            Random seed for reproducibility.
        new: bool
            healpy sinalm keyword which sets the assumed ordering of the Cls. Default: True.
                    If True, use the new ordering of cl’s, ie by diagonal (e.g. TT, EE, BB, TE, EB, TB or TT, EE, BB, TE if 4 cl as input). 
                    If False, use the old ordering, ie by row (e.g. TT, TE, TB, EE, EB, BB or TT, TE, EE, BB if 4 cl as input).

    Returns
    -------
        np.ndarray
            Realization of CMB alms for T, E, and B modes. Shape is (3, n_alms)
    """
    if seed is not None:
        np.random.seed(seed)
    return hp.synalm(cls_cmb, lmax=lmax, new=new)

def _get_foregrounds(
    foreground_models: List[str],
    instrument: dict, 
    nside: int,
    lmax: int,
    return_components: bool = False,
    pixel_window: bool = False,
    units: str = 'uK_CMB',
    return_alms: bool = False,
    bandpass_integrate: bool = False,
    lmin: int = 2,
    coordinates: str = "G",
    **kwargs) -> SimpleNamespace:
    """
    Generate simulated foregrounds from PySM3 models for a given instrument.

    Parameters
    ----------
        foreground_models: (List[str])
            List of PySM3 model presets (e.g., ["d1", "s1"]).
        instrument: dict
            Instrument configuration object with frequency, beams, and optional bandpasses.
        nside: int
            Output HEALPix resolution.
        lmax: int
            Maximum multipole to compute alms.
        return_components: bool, optional
            If True, return individual components instead of just the sum. Default: False.
        pixel_window: bool, optional
            Whether to apply pixel window smoothing. Default: False.
        units: str, optional
            Output units. Default: 'uK_CMB'.
        return_alms: bool, optional
            Whether to return alms instead of maps. Default: False.
        bandpass_integrate: bool, optional
            Whether to integrate foreground components across bandpasses. 
            Default: False (i.e. delta functions are assumed).
        lmin: int, optional
            Minimum multipole to keep (applies filtering).
        coordinates, str, optional
            Target coordinate system for output maps/alms ("G" (Galactic), "E" (Ecliptic), or "C" (Equatorial)). 
            Default: "G"
        **kwargs: Additional keyword arguments for `hp.map2alm`.

    Returns
    -------
        SimpleNamespace: 
            Foregrounds object with `.total` field and optionally individual components.
    """

    nside_ = max(nside, 512)
    
    # Foregrounds initialization
    foregrounds = SimpleNamespace()

    # Derivation of foreground components for all instrument frequencies
    if not return_components or len(foreground_models) == 1:
        sky = pysm3.Sky(nside=nside_, preset_strings=foreground_models, output_unit=getattr(u, units))
        foregrounds.total = _get_foreground_component(
            instrument, sky, nside, lmax,
            pixel_window=pixel_window,
            bandpass_integrate=bandpass_integrate,
            return_alms=return_alms,
            lmin=lmin,
            coordinates=coordinates,
            **kwargs
        )
    else:
        for fmodel in foreground_models:
            sky = pysm3.Sky(nside=nside_, preset_strings=[fmodel], output_unit=getattr(u, units))
            attr = prefix_to_attr.get(fmodel[:3]) or prefix_to_attr.get(fmodel[:2]) or prefix_to_attr.get(fmodel[:1])
            setattr(foregrounds, attr, _get_foreground_component(
                instrument, sky, nside, lmax,
                pixel_window=pixel_window,
                bandpass_integrate=bandpass_integrate,
                return_alms=return_alms,
                lmin=lmin,
                coordinates=coordinates,
                **kwargs
            ))
        foregrounds.total = sum(vars(foregrounds).values())
    return foregrounds

def _get_foreground_component(
    instrument: dict,
    sky: pysm3.Sky,
    nside_out: int,
    lmax: int,
    pixel_window: bool = False,
    bandpass_integrate: bool = False,
    return_alms: bool = False,
    lmin: int = 2,
    coordinates: str = "G",
    **kwargs
) -> np.ndarray:
    """
    Generate a foreground component for each frequency channel of the instrument.

    Parameters
    ----------
        instrument: dict
            Instrument configuration with frequencies, fwhms (or beam paths), bandpasses or bandwidths.
        sky: pysm3.Sky
            PySM3 sky model for the foreground.
        nside_out: int
            HEALPix resolution for the output.
        lmax: int
            Maximum multipole to compute alms.
        pixel_window: bool, optional
            Apply pixel window smoothing if True. Default: False.
        bandpass_integrate: bool, optional
            Integrate over bandpass if True. Default: False.
        return_alms: bool, optional
            Return alms if True, else return maps. Default: False.
        lmin: int, optional
            Minimum multipole to keep (applies filtering).
        coordinates: str, optional
            Coordinate system for output ("G" (Galactic), "E" (Ecliptic), or "C" (Equatorial)). 
            Default: "G"
        **kwargs: Additional arguments passed to `hp.map2alm`.

    Returns
    -------
        np.ndarray: 
            Array of foreground maps or alms with shape (n_channels, 3, npix) for maps or (n_channels, 3, nalms) for alms.
    """

    fg_component = []

    fell = _get_ell_filter(lmin, lmax) if lmin > 2 else None

    rot = hp.Rotator(coord=f"G{coordinates}") if coordinates != "G" else None
    
    # Getting foreground component for each frequency channel
    for idx, freq in enumerate(instrument.frequency):
        if bandpass_integrate:
            if hasattr(instrument, 'path_bandpasses'):
                # Reading bandpass from file
                bandpass_file = instrument.path_bandpasses + f"_{instrument.channels_tags[idx]}.npy"
                frequencies, bandpass_weights = np.load(bandpass_file)
                frequencies = frequencies * u.GHz
            else:
                # Create a top-hat bandpass
                freq_min = freq * (1 - ( instrument.bandwidth[idx] / 2 ))
                freq_max = freq * (1 + ( instrument.bandwidth[idx] / 2 ))
                steps = int(freq_max - freq_min + 1)
                frequencies = np.linspace(freq_min, freq_max, steps) * u.GHz
                bandpass_weights = np.ones(len(frequencies)) # The tophat is defined in intensity units (Jy/sr)
            with threadpool_limits(limits=1):
                emission = sky.get_emission(frequencies, bandpass_weights)
        else:
            with threadpool_limits(limits=1):
                emission = sky.get_emission(freq * u.GHz)

        # Computing alms of the foreground emission
        alm_emission = hp.map2alm(emission.value, lmax=lmax, pol=True, **kwargs)

        # Applying coordinate rotation if needed
        if coordinates != "G":
            rot.rotate_alm(alm_emission, inplace=True)
        
        # Smoothing the alms with the instrument beam
        if instrument.beams == "gaussian":
            alm_emission = _smooth_input_alms_(
                alm_emission,
                fwhm=instrument.fwhm[idx],
                nside_out=nside_out if pixel_window else None
            )
        else:
            beamfile = instrument.path_beams + f"_{instrument.channels_tags[idx]}.fits"
            alm_emission = _smooth_input_alms_(
                alm_emission,
                beam_path=beamfile,
                symmetric_beam=(instrument.beams == "file_l"),
                nside_out=nside_out if pixel_window else None
            )

        if lmin > 2:
            for f in range(3):
                alm_emission[f] = hp.almxfl(alm_emission[f], fell)
        fg_component.append(alm_emission if return_alms else hp.alm2map(alm_emission, nside_out, lmax=lmax, pol=True))
        
    return np.array(fg_component)

def _smooth_input_alms_(
    alms: np.ndarray,
    fwhm: Optional[float] = None,
    nside_out: Optional[int] = None,
    beam_path: Optional[str] = None,
    symmetric_beam: bool = True
) -> np.ndarray:
    """
    Apply beam and pixel window smoothing to input alms.

    Parameters
    ----------
        alms: np.ndarray
            Array of spherical harmonic coefficients [T, E, B].
        fwhm: float, optional
            FWHM of Gaussian beam in arcmin. Required if `beam_path` is None.
        nside_out: int, optional
            HEALPix Nside for pixel window function. Used if not None.
        beam_path: str, optional
            Path to FITS file containing beam transfer functions.
        symmetric_beam: bool, optional
            Whether the beam from FITS file is symmetric (True) or not (False).

    Returns
    -------
        np.ndarray: 
            Smoothed alms.
    """
    lmax = hp.Alm.getlmax(alms.shape[1])

    # Beams transfer functions
    if beam_path is not None:
        bl_i = _get_beam_from_file(beam_path, lmax,symmetric_beam=symmetric_beam)
    elif fwhm is not None:
        symmetric_beam = True
        bl_i = hp.gauss_beam(np.radians(fwhm/60.), lmax = lmax, pol = True)
    else:
        raise ValueError("Either fwhm or beam_path must be provided.")

    bl_i = bl_i[:,:3]  # Take only T, E, B

    if nside_out:
        pw = hp.pixwin(nside_out, pol=True, lmax=lmax)
        pw = np.array([pw[0], pw[1], pw[1]])
        if symmetric_beam:
            bl_i = bl_i * pw.T
        else:
            for i in range(3):
                bl_i[:,i] = hp.almxfl(bl_i[:,i], pw[i])

    # Initializing smoothed alms
    alms_smoothed = np.zeros_like(alms)

    for i in range(3):
        alms_smoothed[i] = hp.almxfl(alms[i], bl_i[:,i]) if symmetric_beam else alms[i] * bl_i[:,i]

    return alms_smoothed

__all__ = [
    name
    for name, obj in globals().items()
    if callable(obj) and getattr(obj, "__module__", None) == __name__
]





from scipy import constants
from astropy.cosmology import Planck18
from pysm3 import units as u
import numpy as np
from pysm3.models.dust import blackbody_ratio
from pysm3.utils import bandpass_unit_conversion, normalize_weights, trapz_step_inplace
from .needlets import _get_needlet_windows_
from typing import List, Union, Optional, Tuple, Dict
import sys

T_cmb = Planck18.Tcmb(0).value
h = constants.h
k = constants.k

def _get_CMB_SED(frequencies: List[float], units: str = "uK_CMB",
    bandwidths: Optional[Union[List[str], List[float]]] = None
) -> np.ndarray:
    """
    Function to compute the SED of the CMB in different units.

    Parameters
    ----------
        frequencies : list
            Frequencies in GHz.
        units : str, optional
            Units of the output SED. Options are "uK_CMB", "uK_RJ" and all their multiples (e.g., "mK_CMB", "mK_RJ") or "Jy_sr".
            Default is "uK_CMB".
        bandwidths : list or None, optional
            If list of str, it should contain the paths to npy files with frequency bands and response.
            If list of floats, it should contain the relative bandwidths in GHz for each frequency channel.
            If None, the SED is computed at the frequencies without assuming any bandwidth.

    Returns
    -------
        np.ndarray
            SED of the CMB in the specified units.
    """

    if units[-3:] == "CMB":
        return (u.Unit(units)).to(u.uK_CMB) * np.ones(len(frequencies))

    if bandwidths is not None:
        sed = []
        for idx_freq, frequency in enumerate(frequencies):
            if isinstance(bandwidths[0], str):
                freqs_band, weights = np.load(bandwidths[idx_freq])
            else:
                freq_min = frequency * (1 - ( bandwidths[idx_freq] / 2 ))
                freq_max = frequency * (1 + ( bandwidths[idx_freq] / 2 ))
                steps = int(freq_max - freq_min + 1)
                freqs_band = np.linspace(freq_min, freq_max, steps)
                weights = np.ones_like(freqs_band) # The tophat is defined in intensity units (Jy/sr)
            weights_rj = normalize_weights(freqs_band, weights)

            sed_band = np.zeros(1)
            for i, (freq, _weight) in enumerate(zip(freqs_band, weights_rj)):
                sed_ = u.uK_CMB.to(u.uK_RJ, equivalencies=u.cmb_equivalencies(freq * u.GHz))
                trapz_step_inplace(freqs_band, weights_rj, i, sed_, sed_band)

            if units != "Jy_sr":
                sed.append((sed_band * bandpass_unit_conversion(freqs_band * u.GHz, weights, getattr(u, units))).value[0])
            else:
                sed.append((sed_band * bandpass_unit_conversion(freqs_band * u.GHz, weights, u.Jy / u.sr)).value[0])

        return np.array(sed)

    else:
        if units[-2:] == "RJ":
            return u.uK_CMB.to(u.Unit(units), equivalencies=u.cmb_equivalencies(frequencies * u.GHz))
        elif units == "Jy_sr":
            return u.uK_CMB.to((u.Jy / u.sr), equivalencies=u.cmb_equivalencies(frequencies * u.GHz))

def _get_moments_SED(
    frequencies: List[float],
    moms_list: List[str],
    beta_d: float = 1.54,
    T_d: float = 20.0,
    beta_s: float = -3.0,
    nu_ref_d: float = 353.0,
    nu_ref_s: float = 30.0,
    units: str = "uK_CMB",
    bandwidths: Optional[Union[List[str], List[float]]] = None
) -> np.ndarray:
    """
    Function to compute the SED of the moments in different units.

    Parameters
    ----------
        frequencies : list
            Frequencies observed in GHz.
        moms_list : list
            List of moments to compute. It can include
            - '0d': zeroth moment of modified black body (thermal dust)
            - '0s': zeroth moment of power law (synchrotron)
            - '1bd': first derivative of modified black body wrt beta
            - '1bs': first derivative of power law wrt beta
            - '1Td': first derivative of modified black body wrt temperature
            - '2bd': second derivative of modified black body wrt beta
            - '2bs': second derivative of power law wrt beta
            - '2Td': second derivative of modified black body wrt temperature
            - '2bdTd': second derivative of modified black body wrt beta and temperature
            - '2Tdbd': second derivative of modified black body wrt temperature and beta
        beta_d : float, optional
            Spectral index of the modified black body (thermal dust). It refers to muK_RJ units SED.
            Default is 1.54.
        T_d : float, optional
            Temperature of the modified black body (thermal dust) in Kelvin. It refers to muK_RJ units SED. 
            Default is 20.
        beta_s : float, optional
            Spectral index of the power law (synchrotron). It refers to muK_RJ units SED. 
            Default is -3.
        nu_ref_d : float, optional
            Reference frequency for the modified black body (thermal dust) in GHz. 
            Default is 353.
        nu_ref_s : float, optional
            Reference frequency for the power law (synchrotron) in GHz.
            Default is 30.
        units : str, optional
            Units of the output SED. Options are "uK_CMB", "uK_RJ", or "Jy_sr" and all their multiples (e.g., "mK_CMB", "mK_RJ", "mJy_sr").
        bandwidths : list or None, optional
            If list of str, it should contain the paths to npy files with frequency bands and response.
            If list of floats, it should contain the relative bandwidths in GHz for each frequency channel.
            If None, the SED is computed at the frequencies without assuming any bandwidth.
    
    Returns
    -------
        np.ndarray
            SED of the moments in the specified units. The shape is (len(moms_list), len(frequencies)).
    """
    
    moments_funcs = {
    '0d': _mbb_0,
    '0s': _pl_0,
    '1bd': _mbb_1b,
    '1bs': _pl_1b,
    '1Td': _mbb_1T,
    '2bd': _mbb_2b,
    '2bs': _pl_2b,
    '2Td': _mbb_2T,
    '2bdTd': _mbb_2bT,
    '2Tdbd': _mbb_2bT,
    }

    all_seds = []
    for mom in moms_list:
        sed = []

        for idx_freq, frequency in enumerate(frequencies):
            if bandwidths is not None:
                if isinstance(bandwidths[0], str):
                    freqs_band, weights = np.load(bandwidths[idx_freq])
                else:
                    freq_min = frequency * (1 - ( bandwidths[idx_freq] / 2 ))
                    freq_max = frequency * (1 + ( bandwidths[idx_freq] / 2 ))
                    steps = int(freq_max - freq_min + 1)
                    freqs_band = np.linspace(freq_min, freq_max, steps)
                    weights = np.ones_like(freqs_band) # The tophat is defined in intensity units (Jy/sr)
                weights_rj = normalize_weights(freqs_band, weights)

                sed_band = np.zeros(1)
                for i, (freq, _weight) in enumerate(zip(freqs_band, weights_rj)):
                    if 'd' in mom:
                        sed_ = moments_funcs[mom](freq, beta_d, T_d, nu_ref = nu_ref_d)
                    elif 's' in mom:
                        sed_ = moments_funcs[mom](freq, beta_s, nu_ref = nu_ref_s)
                    trapz_step_inplace(freqs_band, weights_rj, i, sed_, sed_band)          
            else:
                if 'd' in mom:
                    sed_band = moments_funcs[mom](frequency, beta_d, T_d, nu_ref = nu_ref_d)
                elif 's' in mom:
                    sed_band = moments_funcs[mom](frequency, beta_s, nu_ref = nu_ref_s)
                freqs_band = frequency
                weights = None

            if units != "Jy_sr":
                if units[-3:] == "CMB":
                    sed.append((sed_band * bandpass_unit_conversion(freqs_band * u.GHz, weights, getattr(u, units))).value[0])
                elif units[-2:] == "RJ":
                    sed.append((sed_band * bandpass_unit_conversion(freqs_band * u.GHz, weights, getattr(u, units))).value)
            else:
                sed.append((sed_band * bandpass_unit_conversion(freqs_band * u.GHz, weights, u.Jy / u.sr)).value[0])

        all_seds.append(sed)
    
    return np.array(all_seds)
   

def _mbb_0(nu, beta, T, nu_ref = 335.):
    """
    Function to compute the SED of a modified black body.

    Parameters
    ----------
        nu : float or np.ndarray
            Frequency in GHz.
        beta : float
            Spectral index of the modified black body.
        T : float
            Temperature in Kelvin.
        nu_ref : float, optional
            Reference frequency in GHz. Default is 335 GHz.
    
    Returns
    -------
        np.ndarray
            SED of the modified black body in CMB brightness units.
    """
    x = (h * nu * (1e+9)) / (k * T)
    return ((nu/nu_ref)**(beta+1)) / (np.exp(x)-1)  

def _mbb_1b(nu, beta, T, nu_ref = 335.):
    """
    Function to compute the SED of first derivative of modified black body wrt beta.

    Parameters
    ----------
        nu : float or np.ndarray
            Frequency in GHz.
        beta : float
            Spectral index of the modified black body.
        T : float
            Temperature in Kelvin.
        nu_ref : float, optional
            Reference frequency in GHz. Default is 335 GHz.
        
    Returns
    -------
        np.ndarray
            SED of the first derivative of modified black body wrt beta in CMB brightness units.
    """
    return np.log(nu/nu_ref) * _mbb_0(nu, beta, T, nu_ref = nu_ref)

def _mbb_1T(nu, beta, T, nu_ref = 335.):
    """
    Function to compute the SED of first derivative of modified black body wrt temperature.

    Parameters
    ----------
        nu : float or np.ndarray
            Frequency in GHz.
        beta : float
            Spectral index of the modified black body.
        T : float
            Temperature in Kelvin.
        nu_ref : float, optional
            Reference frequency in GHz. Default is 335 GHz.
    
    Returns
    -------
        np.ndarray
            SED of the first derivative of modified black body wrt temperature in CMB brightness units.
    """
    x = (h * nu * (1e+9)) / (k * T)
    return x * np.exp(x) * _mbb_0(nu, beta, T, nu_ref = nu_ref) / (T * (np.exp(x)-1))

def _mbb_2b(nu, beta, T, nu_ref = 335.):
    """
    Function to compute the SED of second derivative of modified black body wrt beta.

    Parameters
    ----------
        nu : float or np.ndarray
            Frequency in GHz.
        beta : float
            Spectral index of the modified black body.
        T : float
            Temperature in Kelvin.
        nu_ref : float, optional
            Reference frequency in GHz. Default is 335 GHz.

    Returns
    -------
        np.ndarray
            SED of the second derivative of modified black body wrt beta in CMB brightness units.
    """
    return ((np.log(nu/nu_ref))**2) * _mbb_0(nu, beta, T, nu_ref = nu_ref)

def _mbb_2T(nu, beta, T, nu_ref = 335.):
    """
    Function to compute the SED of second derivative of modified black body wrt temperature.

    Parameters
    ----------
        nu : float or np.ndarray
            Frequency in GHz.
        beta : float
            Spectral index of the modified black body.
        T : float
            Temperature in Kelvin.
        nu_ref : float, optional
            Reference frequency in GHz. Default is 335 GHz.

    Returns
    -------
        np.ndarray
            SED of the second derivative of modified black body wrt temperature in CMB brightness units.
    """
    x = (h * nu * (1e+9)) / (k * T)
    return ((x/np.tanh(x/2)) - 2.) * _mbb_1T(nu, beta, T, nu_ref = nu_ref) / T

def _mbb_2bT(nu, beta, T, nu_ref = 335.):
    """
    Function to compute the SED of second derivative of modified black body wrt beta and temperature.

    Parameters
    ----------
        nu : float or np.ndarray
            Frequency in GHz.
        beta : float
            Spectral index of the modified black body.
        T : float
            Temperature in Kelvin.
        nu_ref : float, optional
            Reference frequency in GHz. Default is 335 GHz.

    Returns
    -------
        np.ndarray
            SED of the second derivative of modified black body wrt beta and temperature in CMB brightness units.
    """
    return np.log(nu/nu_ref) * _mbb_1T(nu, beta, T, nu_ref = nu_ref)

def _pl_0(nu, beta, nu_ref = 30.):
    """
    Function to compute the SED of a power law.

    Parameters
    ----------
        nu : float or np.ndarray
            Frequency in GHz.
        beta : float
            Spectral index of the power law.
        nu_ref : float, optional
            Reference frequency in GHz. Default is 30 GHz.

    Returns
    -------
        np.ndarray
            SED of the power law in CMB brightness units.
    """
    return (nu/nu_ref)**(beta)

def _pl_1b(nu, beta, nu_ref = 30.):
    """
    Function to compute the SED of first derivative of power law wrt beta.

    Parameters
    ----------
        nu : float or np.ndarray
            Frequency in GHz.
        beta : float
            Spectral index of the power law.
        nu_ref : float, optional
            Reference frequency in GHz. Default is 30 GHz.

    Returns
    -------
        np.ndarray
            SED of the first derivative of power law wrt beta in CMB brightness units.
    """
    return np.log(nu/nu_ref) * _pl_0(nu, beta, nu_ref = nu_ref)

def _pl_2b(nu, beta, nu_ref = 30.):
    """
    Function to compute the SED of second derivative of power law wrt beta.

    Parameters
    ----------
        nu : float or np.ndarray
            Frequency in GHz.
        beta : float
            Spectral index of the power law.
        nu_ref : float, optional
            Reference frequency in GHz. Default is 30 GHz.

    Returns
    -------
        np.ndarray
            SED of the second derivative of power law wrt beta in CMB brightness units.
    """
    return ((np.log(nu/nu_ref))**2) * _pl_0(nu, beta, nu_ref = nu_ref)

def _standardize_cilc(compsep_run: dict, lmax: int) -> dict:
    """
    Function to standardize and validate the configuration dictionary for cILC component separation.
    It checks and sets the necessary parameters for the cILC method.

    Parameters
    ----------
        compsep_run : dict
            Configuration dictionary for the cILC component separation run.
        lmax : int
            Maximum multipole for the component separation.
    
    Returns
    -------
        compsep_run : dict
            Updated configuration dictionary with standardized parameters.
    """
    if compsep_run["domain"] == "needlet":
        if compsep_run["method"] in ["cilc","cpilc"]:
            nls_number = _get_needlet_windows_(compsep_run["needlet_config"], lmax).shape[0]
        elif compsep_run["method"] in ["c_ilc", "c_pilc"]:
            nls_number = len(compsep_run["special_nls"])
        elif compsep_run["method"] in ["mc_cilc"]:
            nls_number = _get_needlet_windows_(compsep_run["needlet_config"], lmax).shape[0] - len(compsep_run["special_nls"])

    if not "moments" in compsep_run["constraints"]:
        raise ValueError("A list of moments must be provided in the constraints dictionary.")
    else:
        if not isinstance(compsep_run["constraints"]["moments"], list):
            raise ValueError("The moments must be provided as a list.")

    dim_moms = len(compsep_run["constraints"]["moments"]) if isinstance(compsep_run["constraints"]["moments"][0], list) else 1
    if dim_moms == 1 and isinstance(compsep_run["constraints"]["moments"][0], list):
        compsep_run["constraints"]["moments"] = compsep_run["constraints"]["moments"][0]

    if (dim_moms > 1) and (compsep_run["domain"]=="pixel"):
        raise ValueError("If the domain is pixel, moments must be a 1d list.")

    if "deprojection" in compsep_run["constraints"]:
        if not isinstance(compsep_run["constraints"]["deprojection"], (int, float, list)):
            raise ValueError("The deprojection parameter must be a scalar or a list.")
        if isinstance(compsep_run["constraints"]["deprojection"], list):
            dim_depro = len(compsep_run["constraints"]["deprojection"]) if isinstance(compsep_run["constraints"]["deprojection"][0], list) else 1
            if dim_depro != dim_moms:
                raise ValueError("The list of deprojection coefficients must have the same dimension as the moments list.")

            if dim_depro == 1 and isinstance(compsep_run["constraints"]["deprojection"][0], list):
                compsep_run["constraints"]["deprojection"] = compsep_run["constraints"]["deprojection"][0]
            
            if dim_depro == 1:
                if len(compsep_run["constraints"]["deprojection"]) != len(compsep_run["constraints"]["moments"]):
                    raise ValueError("The list of deprojection coefficients must have the same dimension as the moments list.")
            else:
                for n in range(len(compsep_run["constraints"]["deprojection"])):
                    if len(compsep_run["constraints"]["deprojection"][n]) != len(compsep_run["constraints"]["moments"][n]):
                        raise ValueError("The list of deprojection coefficients must have the same dimension as the moments list.")
            
    if compsep_run["domain"] == "needlet":
        if dim_moms > 1:
            if dim_moms > nls_number:
                compsep_run["constraints"]["moments"] = compsep_run["constraints"]["moments"][:nls_number]
            elif dim_moms < nls_number:
                while len(compsep_run["constraints"]["moments"]) < nls_number:
                    (compsep_run["constraints"]["moments"]).append(compsep_run["constraints"]["moments"][-1])
        else:
            compsep_run["constraints"]["moments"] = [compsep_run["constraints"]["moments"] for _ in range(nls_number)]

    if not "deprojection" in compsep_run["constraints"]:
        if compsep_run["domain"] == "pixel":
            compsep_run["constraints"]["deprojection"] = [0. for _ in compsep_run["constraints"]["moments"]]
        elif compsep_run["domain"] == "needlet":
            compsep_run["constraints"]["deprojection"] = [[0. for _ in row] for row in compsep_run["constraints"]["moments"]]
    elif isinstance(compsep_run["constraints"]["deprojection"], (int, float)):
        if compsep_run["domain"] == "pixel":
            compsep_run["constraints"]["deprojection"] = [compsep_run["constraints"]["deprojection"] for _ in compsep_run["constraints"]["moments"]]
        elif compsep_run["domain"] == "needlet":
            compsep_run["constraints"]["deprojection"] = [[compsep_run["constraints"]["deprojection"] for _ in row] for row in compsep_run["constraints"]["moments"]]
    elif isinstance(compsep_run["constraints"]["deprojection"], list):
        if compsep_run["domain"] == "needlet":
            if dim_depro > 1:
                if dim_depro > nls_number:
                    compsep_run["constraints"]["deprojection"] = compsep_run["constraints"]["deprojection"][:nls_number]
                elif dim_depro < nls_number:
                    while len(compsep_run["constraints"]["deprojection"]) < nls_number:
                        (compsep_run["constraints"]["deprojection"]).append(compsep_run["constraints"]["deprojection"][-1])
            else:
                compsep_run["constraints"]["deprojection"] = [compsep_run["constraints"]["deprojection"] for _ in range(nls_number)]

    if not "beta_d" in compsep_run["constraints"]:
        compsep_run["constraints"]["beta_d"] = 1.54
    
    if not isinstance(compsep_run["constraints"]["beta_d"], (int, float, list)):
        raise ValueError("The beta_d parameter must be a scalar or a list.")
    else:
        if isinstance(compsep_run["constraints"]["beta_d"], list):
            dim_beta = len(compsep_run["constraints"]["beta_d"]) if isinstance(compsep_run["constraints"]["beta_d"][0], list) else 1
            if dim_beta == 1 and isinstance(compsep_run["constraints"]["beta_d"][0], list):
                compsep_run["constraints"]["beta_d"] = compsep_run["constraints"]["beta_d"][0]
            elif dim_beta > 1:
                raise ValueError("The beta_d parameter must be a scalar or a list of scalars.")

        if compsep_run["domain"] == "needlet":
            if isinstance(compsep_run["constraints"]["beta_d"], (int, float)):
                compsep_run["constraints"]["beta_d"] = [compsep_run["constraints"]["beta_d"] for _ in range(nls_number)]
            elif len(compsep_run["constraints"]["beta_d"]) < nls_number:
                while len(compsep_run["constraints"]["beta_d"]) < nls_number:
                    (compsep_run["constraints"]["beta_d"]).append(compsep_run["constraints"]["beta_d"][-1])
            elif len(compsep_run["constraints"]["beta_d"]) > nls_number:
                compsep_run["constraints"]["beta_d"] = compsep_run["constraints"]["beta_d"][:nls_number]
        elif compsep_run["domain"] == "pixel":
            if isinstance(compsep_run["constraints"]["beta_d"], list):
                raise ValueError("If the domain is pixel, beta_d must be a scalar.")

    if not "T_d" in compsep_run["constraints"]:
        compsep_run["constraints"]["T_d"] = 20.
    
    if not isinstance(compsep_run["constraints"]["T_d"], (int, float, list)):
        raise ValueError("The T_d parameter must be a scalar or a list.")
    else:
        if isinstance(compsep_run["constraints"]["T_d"], list):
            dim_T = len(compsep_run["constraints"]["T_d"]) if isinstance(compsep_run["constraints"]["T_d"][0], list) else 1
            if dim_T == 1 and isinstance(compsep_run["constraints"]["T_d"][0], list):
                compsep_run["constraints"]["T_d"] = compsep_run["constraints"]["T_d"][0]
            elif dim_T > 1:
                raise ValueError("The T_d parameter must be a scalar or a list of scalars.")

        if compsep_run["domain"] == "needlet":
            if isinstance(compsep_run["constraints"]["T_d"], (int, float)):
                compsep_run["constraints"]["T_d"] = [compsep_run["constraints"]["T_d"] for _ in range(nls_number)]
            elif len(compsep_run["constraints"]["T_d"]) < nls_number:
                while len(compsep_run["constraints"]["T_d"]) < nls_number:
                    (compsep_run["constraints"]["T_d"]).append(compsep_run["constraints"]["T_d"][-1])
            elif len(compsep_run["constraints"]["T_d"]) > nls_number:
                compsep_run["constraints"]["T_d"] = compsep_run["constraints"]["T_d"][:nls_number]
        elif compsep_run["domain"] == "pixel":
            if isinstance(compsep_run["constraints"]["T_d"], list):
                raise ValueError("If the domain is pixel, T_d must be a scalar.")
                
    if not "beta_s" in compsep_run["constraints"]:
        compsep_run["constraints"]["beta_s"] = -3.
    
    if not isinstance(compsep_run["constraints"]["beta_s"], (int, float, list)):
        raise ValueError("The beta_s parameter must be a scalar or a list.")
    else:
        if isinstance(compsep_run["constraints"]["beta_s"], list):
            dim_beta = len(compsep_run["constraints"]["beta_s"]) if isinstance(compsep_run["constraints"]["beta_s"][0], list) else 1
            if dim_beta == 1 and isinstance(compsep_run["constraints"]["beta_s"][0], list):
                compsep_run["constraints"]["beta_s"] = compsep_run["constraints"]["beta_s"][0]
            elif dim_beta > 1:
                raise ValueError("The beta_s parameter must be a scalar or a list of scalars.")

        if compsep_run["domain"] == "needlet":
            if isinstance(compsep_run["constraints"]["beta_s"], (int, float)):
                compsep_run["constraints"]["beta_s"] = [compsep_run["constraints"]["beta_s"] for _ in range(nls_number)]
            elif len(compsep_run["constraints"]["beta_s"]) < nls_number:
                while len(compsep_run["constraints"]["beta_s"]) < nls_number:
                    (compsep_run["constraints"]["beta_s"]).append(compsep_run["constraints"]["beta_s"][-1])
            elif len(compsep_run["constraints"]["beta_s"]) > nls_number:
                compsep_run["constraints"]["beta_s"] = compsep_run["constraints"]["beta_s"][:nls_number]
        elif compsep_run["domain"] == "pixel":
            if isinstance(compsep_run["constraints"]["beta_s"], list):
                raise ValueError("If the domain is pixel, beta_s must be a scalar.")

    return compsep_run

__all__ = [
    name
    for name, obj in globals().items()
    if callable(obj) and getattr(obj, "__module__", None) == __name__
]
